from __future__ import print_function

from kabaret import flow

from .action_context import ActionContextGroup
from .action_submenus_demo import ActionSubmenusDemo
from .action_dialog_size import DialogSizeOptions

class NoDialogAction(flow.Action):

    def needs_dialog(self):
        return False

    def run(self, button):
        print('Well done Jolly Jumper !')

class SimpleDialogAction(flow.Action):

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is the action message: <3')
        return ['Ok', 'Maybe', 'Nope']

    def run(self, button):
        print('You chose {}'.format(button))

class DialogParamsAction(flow.Action):

    cut_down_trees = flow.BoolParam(True)
    skip_and_jump = flow.BoolParam(True)
    press_wild_flowers = flow.BoolParam(True)

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Do you:')
        return ['Ok']

    def run(self, button):
        all_true = (
            self.cut_down_trees.get()
            and self.skip_and_jump.get()
            and self.press_wild_flowers.get()
        )
        if (
            self.cut_down_trees.get()
            and self.skip_and_jump.get()
            and self.press_wild_flowers.get()
        ):
            print('You hang around in bars')
        else:
            print('Do you ever party ?')

class DispatchParams(flow.Object):

    pool = flow.Param('Farm')
    priority = flow.Param(50)

    def get_flags(self):
        return [
            '-P', str(self.pool.get()),
            '-p', str(self.priority.get()),
        ]


class SequenceParam(flow.Object):

    first_frame = flow.Param(1)
    last_frame = flow.Param(100)

    def get_flags(self):
        return [
            '--first', str(self.first_frame.get()),
            '--last', str(self.last_frame.get()),
        ]

class ComplexDialogAction(flow.Action):

    dispatch = flow.Child(DispatchParams)
    sequence = flow.Child(SequenceParam)

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Configure and Submit your job')
        return ['Submit']

    def run(self, button):
        cmd = ['spam_it']+self.dispatch.get_flags()+self.sequence.get_flags()
        print('#---> Cmd:', ' '.join(cmd))

class KeepOpenDialogAction(flow.Action):

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Click the Ok button !')
        return ['Ok', 'Close']

    def run(self, button):
        if button == 'Close':
            return

        self.message.set('Alrigth, now click the Close button :)')
        return self.get_result(close=False)

class GotoAction(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self._target_name = None
        return ['Goto Project', 'Goto Parent', 'Create Page and use it for all Goto', 'Open new a page every time',]

    def run(self, button):
        if button == 'Goto Project':
            project_oid = self.root().project().oid()
            return self.get_result(goto=project_oid, close=False, goto_target=self._target_name)

        if button == 'Goto Parent':
            parent_oid = self._parent.oid()
            return self.get_result(goto=parent_oid, close=False, goto_target=self._target_name)

        if button == 'Create Page and use it for all Goto':
            project_oid = self.root().project().oid()
            import time
            self._target_name = str(time.time())
            return self.get_result(goto=project_oid, close=False, goto_target=self._target_name)

        if button == 'Open new a page every time':
            project_oid = self.root().project().oid()
            return self.get_result(goto=project_oid, close=False, goto_target='_NEW_')

class GotoActionTitleFormater(flow.Action):
     
    _parent = flow.Parent()

    formater = flow.Param('-=# {} #=-')

    def needs_dialog(self):
        return True
    
    def get_buttons(self):
        self.message.set(
            '''
            <html>
            You can control the title of a view created 
            for an Action goto.<br>
            Enter a text containing "{}" in the formater
            field and click "Open".
            </html>
            '''
        )
        return ['Open', 'Close']
    
    def run(self, button):
        if button != 'Open':
            return 

        # You specify the format to use by building a view ID
        # in the form: <formater>|<view_id>
        # See the goto_target argument here:
        return self.get_result(
            goto=self.root().project().oid(),
            goto_target=str(self.formater.get())+'|VIEW_TITLW_SHOWCASE'
        )

class EditMyValueAction(flow.Action):

    _value = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Pick a color:')
        return ['Red', 'Pink', 'Blue', 'Violet']

    def run(self, button):
        self._value.set(button)

class SetMyValueAction(flow.Action):

    _value = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        self._value.set('orange')

class MyValue(flow.values.Value):

    edit = flow.Child(EditMyValueAction)
    orange = flow.Child(SetMyValueAction)

class SubDialogAction(flow.Action):

    quote = flow.Param('', MyValue)

    def needs_dialog(self):
        return True

    def run(self, button):
        print('You selected: "{}"'.format(self.quote.get()))

class WizardDialogPage1(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is page <B>1/3</B> of a wizard style action')
        return ['Next']

    def run(self, button):
        return self.get_result(next_action=self._parent.wizard_page_2.oid())

class WizardDialogPage2(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is page <B>2/3</B> of a wizard style action')
        return ['Prev', 'Next']

    def run(self, button):
        if button == 'Next':
            return self.get_result(next_action=self._parent.wizard_page_3.oid())
        else:
            return self.get_result(next_action=self._parent.wizard_page_1.oid())

class WizardDialogPage3(flow.Action):

    _parent = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('This is page <B>3/3</B> of a wizard style action')
        return ['Prev', 'Finish']

    def run(self, button):
        if button == 'Prev':
            return self.get_result(next_action=self._parent.wizard_page_2.oid())
        print('#---> You are a Wizard now !!!')


class ActionsGroup(flow.Object):

    no_dialog_action = flow.Child(NoDialogAction)
    simple_dialog_action = flow.Child(SimpleDialogAction)
    keep_dialog_open_action = flow.Child(KeepOpenDialogAction)
    goto_action = flow.Child(GotoAction)
    goto_title_formater = flow.Child(GotoActionTitleFormater)
    dialog_params_action = flow.Child(DialogParamsAction)
    complex_dialog_action = flow.Child(ComplexDialogAction)

    sub_dialog_action = flow.Child(SubDialogAction)

    wizard_page_1 = flow.Child(WizardDialogPage1)
    wizard_page_2 = flow.Child(WizardDialogPage2).ui(hidden=True)
    wizard_page_3 = flow.Child(WizardDialogPage3).ui(hidden=True)

    action_submenu_demo = flow.Child(ActionSubmenusDemo)

    sep = flow.Label(
        '<hr>'
        '<h2>Action Context</h2>'
        'Action are accessible from various places in the GUI. '
        'Each Action class can specify if it wants to appear in a given context by '
        'implementing the <b>allow_context(context)</b> method. Returning True will let '
        'the GUI show the action, returning False will prevent the GUI from considering '
        'the action. Note that even if shown, the action may be disabled by calling '
        'ui(enabled=False) on the relation pointing to it with<br> '
        'The context argument value will depend of the GUI context gathering the actions: '
        '<ul>'
        '<li>'
            'In Flow view\'s "inline" object representation:<br> '
            'A down arrow button is show at the right of Objects. '
            'It shows every Child action of the Object. '
            'This context is <font color=orange><B>Flow.inline</B></font><br>'
            'Note that this menu may in some case show actions from grand-children. '
            'In such cases, the context has a number of ">" appended: '
            '<font color=orange><B>Flow.inline&gt;&gt;&gt;</B></font><br>'
            'The number of ">" corresponds to the depth from the Object showing '
            'the inline representation.'
        '</li>'
        '<li>'
            'In Flow view\'s "details" object representation:<br> '
            'A button is show inside Objects for each of its '
            'Children actions. '
            'This context is <font color=orange><B>Flow.details</B></font> '
        '</li>'
        '<li>'
            'In Flow view\'s mapped items representation:<br> '
            'A right-click shows an action menu. '
            'This context is <font color=orange><B>Flow.map</B></font> '
        '</li>'
        '<li>'
            'Other view\'s may display action. Most views will use a context '
            'named after the view_type. For example, the ScriptView uses the '
            '<font color=orange><b>Script</b></font> context.'
        '</li>'
        '<li>'
            'The context may be <font color=orange><b>None</b></font> with legacy view '
            'code. In this case you can choose to return None so that the legacy '
            'behavior is used (using the class <i>SHOW_IN_PARENT_INLINE</i> and '
            '<i>SHOW_IN_PARENT_DETAILS</i> attributes) '
            'You should not relay on this since it will be deprecated at some point. '
        '</li>'
        '</ul>'
        'Enter the group below to explore a set of examples:'
    )
    action_context = flow.Child(ActionContextGroup)

    sep_2 = flow.Label(
        """
        <hr>
        <h2>Action Dialog Size</h2>
        <p>
            The size of an Action dialog at display can be configured using the <b>dialog_size</b> UI option.
            A minimum and a maximum size can also be specified using the <b>dialog_min_size</b> and
            <b>dialog_max_size</b> options. A size must be a tuple of the form (<i>width</i>, <i>height</i>).<br><br>
            
            In case the default size is outside the provided range, it is reset to the nearest limit. Invalid
            size limits are ignored though, in which case the default size is left unchanged.
            Providing a negative size (e.g., (-1, -1)) disables the option. It can be useful to disable
            a size option provided in a parent class of the Action.
            </p>
            <p>
            These options can be used either in the Action relation's <tt>ui()</tt> method
            <pre>>   Relation.ui(dialog[_min|_max]_size=(width, height))</pre>
            or in the Action's <tt>_fill_ui()</tt> method
            <pre>>   def _fill_ui(self, ui):
        ui['dialog[_min|_max]_size'] = (width, height)</pre>
            </p>
        </p>
        """
    )
    action_dialog_size_options = flow.Child(DialogSizeOptions)


'''

    This flow defines a generic "Home" page for Flow Actor.

'''
from __future__ import print_function

import six

from kabaret import flow

from .utils import import_object



class ProjectNameChoiceValue(flow.values.ChoiceValue):

    def choices(self):
        projects_info = self.root().flow_actor.get_projects_info()
        return [n for n, i in projects_info]


class GetProjectTypeFromProjectAction(flow.Action):

    _ptv = flow.Parent()

    project = flow.Param('', ProjectNameChoiceValue)

    def needs_dialog(self):
        return True

    def get_buttons(self):
        return ['Get Project Type']

    def run(self, button):
        projects_info = dict(self.root().flow_actor.get_projects_info())
        project_info = projects_info.get(self.project.get())
        if not project_info:
            self.message.set('Unknow Project %r' % (self.project.get()))
            return self.get_result(close=False)
        project_type = project_type['type']
        self._ptv.set(project_type)


class ProjectTypeValue(flow.values.Value):

    from_project = flow.Child(GetProjectTypeFromProjectAction)


class CreateProjectAction(flow.Action):

    ICON = ('icons.gui', 'plus-sign-in-a-black-circle')

    _projects = flow.Parent()
    project_name = flow.Param('MyProject')
    project_type = flow.Param(
        'dev_studio.flow.demo_project.DemoProject',
        #ProjectTypeValue
        # NB: we can't use this value bc of a bug in action dialog management
        # that prevents me from having an action dialog inside an action 
        # dialog. We'll activate this once the bug is fixed
        # (Also: remember to update the tutorials text & screenshots !!!)
    )

    def get_buttons(self):
        return ['Test Type', 'Create Project']

    def run(self, button):
        flow_actor = self.root().flow_actor

        if button == 'Test Type':
            try:
                TYPE = import_object(self.project_type.get())
            except Exception as err:
                self.message.set(
                    '<font color=red>Error:</font><br>%s' % (err,)
                )
            else:
                self.message.set('Project Type looks good:\n%s' % (TYPE,))
            return self.get_result(close=False)

        else:
            flow_actor.create_project(
                self.project_name.get(),
                self.project_type.get()
            )
            self._projects.touch()
            return self.get_result()


class ProjectStatusChoiceValue(flow.values.ChoiceValue):

    CHOICES = ('NYS', 'WIP', 'DONE', 'Archived')


class SetProjectStatusAction(flow.Action):

    ICON = 'input'

    _home = flow.Parent()

    #project_name = flow.Param('')
    project = flow.Param('', ProjectNameChoiceValue)
    status = flow.Param('WIP', ProjectStatusChoiceValue)

    def get_buttons(self):
        return ['Set Status']

    def run(self, button):
        actor = self.root().flow_actor
        project_name = self.project.get().strip()
        if not project_name:
            self.message.set('Please select a project !')
            return self.get_result(close=False)

        actor.set_project_status(project_name, self.status.get())
        self._home.touch()


class ToggleArchivedProjectsAction(flow.Action):

    _home = flow.Parent()

    def needs_dialog(self):
        return False

    def get_buttons(self):
        return []

    def run(self, button):
        self._home.do_toggle_archived_projects()


class ToggleProjectsTypeAction(flow.Action):

    _home = flow.Parent()

    def needs_dialog(self):
        return False

    def get_buttons(self):
        return []

    def run(self, button):
        self._home.do_toggle_projects_type()


class ProjectsMap(flow.Map):

    ICON = 'asset_family'

    toggle_archived_projects = flow.Child(ToggleArchivedProjectsAction)
    toggle_project_type = flow.Child(ToggleProjectsTypeAction)

    create_project = flow.Child(CreateProjectAction).ui(group='Admin')
    set_project_status = flow.Child(SetProjectStatusAction).ui(group='Admin')

    def __init__(self, *args, **kwargs):
        super(ProjectsMap, self).__init__(*args, **kwargs)
        self._show_archived = False
        self._show_project_type = False

    def do_toggle_archived_projects(self):
        self._show_archived = not self._show_archived
        self.touch()

    def do_toggle_projects_type(self):
        self._show_project_type = not self._show_project_type
        self.touch()

    def columns(self):
        cols = ['Name', 'Status']
        if self._show_project_type:
            cols.append('Type')
        return cols

    def rows(self):
        rows = []
        projects_info = self.root().flow_actor.get_projects_info()
        for name, infos in projects_info:
            type_name = infos['type']
            status = infos['status']
            if not self._show_archived and status == 'Archived':
                continue
            style = dict(
                icon=('icons.gui', 'team'),
                Status_icon=('icons.status', status),
            )
            rows.append((
                '/' + name,
                dict(
                    Name=name,
                    Status=status,
                    Type=type_name,
                    _style=style,
                )
            ))
        return rows


class Home(flow.SessionObject):

    projects = flow.Child(ProjectsMap).ui(
        auto_fit=False, columns_width=(50, 20))


class AbstractHomeRoot(flow.Root):
    '''
    This is the base class for any Home flow root.

    In order to use a customg Home page, one must subclass this, 
    add a 'Home' relation, and pass it as the CustomHomeRootType
    argument of the Flow actor constructor in
    KabaretSession()._create_actors().

    '''

    def set_flow_actor(self, flow_actor):
        self.flow_actor = flow_actor


class HomeRoot(AbstractHomeRoot):

    Home = flow.Child(Home)




from warnings import warn

from .interface import Editor_Interface

from qtpy import QtWidgets, QtGui, QtCore

from ..event_filters import IgnoreMouseButton


class MultiChoiceValueEditor(QtWidgets.QListWidget, Editor_Interface):
    '''
    This editor lets you choose several items in a choice list.

    Editor Type Names:
        multichoice
        multichoices    (deprecated)
        multi_choices   (deprecated)

    Options:
        sorted:         bool (False) - sorts the choices
        default_height : int  (None)  - Force default width

    Deprecated options from previous kabaret versions:
        allow_add:  You must use child actions to provide this fonctionality
    '''
    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        if editor_type_name in ('multichoices', 'multi_choices'):
            warn('Editor Type Name {!r} is deprecated.'.format(
                editor_type_name))
        return editor_type_name in ('multichoice', 'multichoices', 'multi_choices')

    def __init__(self, parent, options):
        QtWidgets.QListWidget.__init__(self, parent)
        Editor_Interface.__init__(self, parent)
        self.viewport().installEventFilter(IgnoreMouseButton(self))
        self.installEventFilter(IgnoreMouseButton(self))

        self._editable = True
        self._sorted = False
        self._default_height = None

        self.apply_options(options)

        self._current_value = []

    def needs_choices(self):
        '''
        Must be overriden by editor presenting a choice of possible values.
        '''
        return True

    def set_editable(self, b):
        '''
        Must be implemented to prevent editing if b is False.
        Visual cue show also be given to the user.
        '''
        self.setEnabled(b)
        if b:
            self.setSelectionMode(self.ExtendedSelection)
            self.itemActivated.connect(self._on_item_activate)
            self.itemChanged.connect(self._on_item_changed)

    def apply_options(self, options):
        '''
        Must be implemented to configure the editor as
        described by the options dict.
        '''
        if 'allow_add' in options:
            raise ValueError("Obsolete option 'allow_add'")

        self._default_height = options.get('default_height')
        self._sorted = options.get('sorted', False)

    def update(self):
        '''
        Must be implemnented to show the value returned by self.fetch_value()
        Your code should call self._on_updated() at the end.
        '''
        self.blockSignals(True)
        # this is because needs_choices() returns True
        value, choices = self.fetch_value()
        self._current_value = value     # avoid copy

        self.clear()
        try:
            remaining = set(self._current_value)
        except TypeError as err:
            self._show_error(str(err))
            return

        for choice in choices or []:
            if choice is None:
                label = ''
            else:
                label = str(choice)
            item = QtWidgets.QListWidgetItem(label, self)
            item._choice = choice

            if not self._editable:
                item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEnabled)

            if choice in self._current_value:
                item.setCheckState(QtCore.Qt.Checked)
            else:
                item.setCheckState(QtCore.Qt.Unchecked)

            icon = self.get_icon_for(label)
            if icon is not None:
                item.setIcon(icon)

            try:
                remaining.remove(choice)
            except KeyError:
                pass

        for choice in remaining:
            if choice is None:
                label = ''
            else:
                label = str(choice)
            item = QtWidgets.QListWidgetItem(label, self)
            item._choice = choice

            if not self._editable:
                item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEnabled)

            item.setCheckState(QtCore.Qt.Checked)

            icon = self.get_icon_for(label)
            if icon is not None:
                item.setIcon(icon)

        if self._sorted:
            self.sortItems()

        self.setMinimumSize(self.minimumSizeHint())

        self.blockSignals(False)
        self._on_updated()

    def minimumSizeHint(self):
        if self._default_height is not None:
            return QtCore.QSize(self.width(), self._default_height)

        if self.model().rowCount() == 0:
            return QtCore.QSize(self.width(), 0)

        height = (self.model().rowCount() + 0.2) * self.sizeHintForRow(0)
        return QtCore.QSize(self.sizeHintForColumn(0), min(height, 400))

    def sizeHint(self):
        if self._default_height is not None:
            return QtCore.QSize(self.width(), self._default_height)

        if self.model().rowCount() == 0:
            return QtCore.QSize(self.width(), 0)

        height = (self.model().rowCount() + 0.4) * self.sizeHintForRow(0)
        return QtCore.QSize(self.sizeHintForColumn(0), min(height, 500))

    def _on_item_activate(self, item):
        checked = item.checkState()
        if checked != QtCore.Qt.Checked:
            item.setCheckState(QtCore.Qt.Checked)
        else:
            item.setCheckState(QtCore.Qt.Unchecked)

    def _on_item_changed(self, item):
        if self.selectedItems():
            items = [item._choice for item in self.selectedItems()]
        else:
            items = [item._choice]

        if item.checkState() & QtCore.Qt.Checked:
            self._current_value = self._current_value + items
        else:
            for item in items:
                self._current_value.remove(item)
        self.apply()

    def get_edited_value(self):
        '''
        Must be implemented to return the value currently displayed.
        '''
        return self._current_value

    def _show_edited(self):
        '''
        Must be implemented to show that the displayed value
        needs to be applied.
        '''
        # This will not happend, the changes are directly applied.
        self.setProperty('edited', True)
        self.setProperty('applying', False)
        self.style().polish(self)

    def _show_applied(self):
        '''
        Must be implemented to show that the displayed value
        as been saved.
        In a clean scenario, applying edits will trigger an update()
        and this state should disapear.
        If you are using the Editor without this kind of round trip,
        you can call update here.
        '''
        self.setProperty('applying', True)

    def _show_clean(self):
        '''
        Must be implemented to show that the displayed value is 
        up to date.
        '''
        self.setProperty('edited', False)
        self.setProperty('applying', False)
        self.style().polish(self)

    def _show_error(self, error_message):
        '''
        Must be implemented to show that the given error occured.
        '''
        self.setProperty('error', True)
        self.style().polish(self)
        msg = '/!\\ ERROR: ' + error_message
        self.setToolTip(self.toolTip() + '\n' + msg)



'''


    Maya User Setup File.

    The PYTHONPATH must be set to the dir of this file for maya to
    exec it at startup.

    BEWARE:
        MAYA WILL NOT REPORT ERRORS HERE. 
        YOU MUST CATCH AND REPORT EXCEPTIONS YOURSELF!

'''
import os, sys, traceback
import logging

if __name__ == '__main__':

    try:
        kabaret_home = os.environ['KABARET_INSTALL']
    except KeyError:
        logging.getLogger('kabaret.demo').warning('Could not find KABARET_HOME in env :(')
    else:
        try:
            dev_studio_home = os.environ['DEV_STUDIO_INSTALL']
        except KeyError:
            logging.getLogger('kabaret.demo').warning('Could not find DEV_STUDIO_HOME in env :(')
        else:
            logging.getLogger('kabaret.demo').info('Installing Kabaret from ' + kabaret_home)
            if kabaret_home not in sys.path:
                sys.path.append(kabaret_home)
            try:
                import kabaret
            except ImportError:
                traceback.print_exc()
                logging.getLogger('kabaret.demo').info('=========> !!! Could not import kabaret package !!!')
            else:
                logging.getLogger('kabaret.demo').info('   Ok.')

            logging.getLogger('kabaret.demo').info('Installing Dev Studio from ' + dev_studio_home)
            if dev_studio_home not in sys.path:
                sys.path.append(dev_studio_home)
            try:
                import dev_studio
            except ImportError:
                traceback.print_exc()
                logging.getLogger('kabaret.demo').info('=========> !!! Could not import dev_studio package !!!')
            else:
                logging.getLogger('kabaret.demo').info('   Ok.')

    # From now on, print will fall into the Script Editor:
    import maya.utils
    maya.utils.executeDeferred('import dev_studio.flow.demo_project.maya_startup as ms; ms.install_kabaret()')



from qtpy import QtWidgets, QtGui, QtCore


class PopupMenuItem(QtWidgets.QListWidgetItem):

    def __init__(self, text, action=None):
        super(PopupMenuItem, self).__init__(text)
        self._action = action


class PopupMenuListWidget(QtWidgets.QListWidget):

    def sizeHint(self):
        s = QtCore.QSize()
        margins = self.contentsMargins()
        count = self.count()
        list_size = count * (self.sizeHintForRow(0)+self.spacing()*2) + margins.top() + margins.bottom()

        # Cannot explain why I add more spacing but it's needed for avoiding unnecessary scrollbar
        s.setHeight(list_size+self.spacing()*2)
        s.setWidth(self.sizeHintForColumn(0) + margins.left() + margins.right())
        return s


class PopupMenu(QtWidgets.QDialog):

    def __init__(self, parent):
        super(PopupMenu, self).__init__(parent)
        self.setWindowFlags(QtCore.Qt.Popup)
        layout = QtWidgets.QVBoxLayout(self)
        layout.setContentsMargins(15, 3, 5, 3)
        layout.setSpacing(0)

        palette = self.palette()
        palette.setColor(QtGui.QPalette.Base, palette.color(QtGui.QPalette.Window))
        self.setPalette(palette)

        self.setLayout(layout)

        self._lists = []
        self._add_new_list()

    def addAction(self, text, icon=None, callback=None):
        action = QtWidgets.QAction(self)
        item = PopupMenuItem(text, action)
        item.setData(QtCore.Qt.UserRole, text)
        self._lists[-1].addItem(item)
        if callback:
            action.triggered.connect(callback)
        if icon:
            item.setIcon(icon)
        return action

    def addItem(self, text, icon=None):
        item = PopupMenuItem(text)
        item.setData(QtCore.Qt.UserRole, text)
        self._lists[-1].addItem(item)
        if icon:
            item.setIcon(icon)
        return item

    def addSeparator(self):
        if not self._lists[-1].count():
            return
        list = self._lists[-1]
        item = QtWidgets.QListWidgetItem()
        item.setFlags(QtCore.Qt.NoItemFlags)
        frame = QtWidgets.QFrame(list)
        frame.setFrameStyle(QtWidgets.QFrame.HLine | QtWidgets.QFrame.Plain)
        list.addItem(item)
        list.setItemWidget(item, frame)

    def addMenu(self):
        if not self._lists[-1].count():
            return
        frame = QtWidgets.QFrame(self)
        frame.setFrameStyle(QtWidgets.QFrame.HLine | QtWidgets.QFrame.Plain)
        self.layout().addWidget(frame)
        self._add_new_list()

    def _add_new_list(self):
        list = PopupMenuListWidget(self)
        list.setSpacing(3)
        list.setFrameStyle(QtWidgets.QFrame.NoFrame)
        list.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        list.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        list.itemClicked.connect(self._onItemClicked)

        self._lists.append(list)
        self.layout().addWidget(list, 1)

    def clear(self):
        layout = self.layout()
        child = layout.takeAt(0)
        while child:
            widget = child.widget()
            if isinstance(widget, QtWidgets.QListWidget):
                child.widget().clear()
            child = layout.takeAt(0)
            widget.deleteLater()
        self._lists = []
        self._add_new_list()

    def popup(self, pos):
        for list in self._lists:
            size = list.sizeHint()
            list.setMinimumWidth(size.width())
            list.setMaximumHeight(size.height())
        self.move(pos)
        self.show()
        self.adjustSize()

    def _onItemClicked(self, item):
        if item._action:
            item._action.trigger()
            self.setVisible(False)



from qtpy import QtWidgets, QtCore

from .python_value import PythonValueEditor


class PasswordEditor(PythonValueEditor):
    '''
    This editor lets you enter password text value.

    Editor Type Names:
        password
    '''

    @classmethod
    def can_edit(cls, editor_type_name):
        '''
        Must be implemented to return True if the given editor_type_name
        matches this editor.
        '''
        return editor_type_name in ('password',)

    def __init__(self, parent, options):
        super(PasswordEditor, self).__init__(parent, options)
        self.apply_options(options)
        self.setEchoMode(QtWidgets.QLineEdit.Password)
        self.setInputMethodHints(QtCore.Qt.ImhHiddenText | QtCore.Qt.ImhNoPredictiveText | QtCore.Qt.ImhNoAutoUppercase)




from kabaret import flow
from kabaret.app.actors.flow.generic_home_flow import AbstractHomeRoot


class Home(flow.Object):

    congrat = flow.Param('You made your own Home !').ui(editable=False)


class MyHomeRoot(AbstractHomeRoot):
    '''
    '''

    Home = flow.Child(Home)

    def set_flow_actor(self, flow_actor):
        self.flow_actor = flow_actor



import os
import traceback
import six

from qtpy import QtWidgets, QtGui, QtCore # those import are also to ease custom page dev

from ...icons import flow  # noqa # to be sure flow icons are registered.

# !!from ..view import View
from ..widget_view import DockedView, DialogView

from .. import event_filters

from .flow_form import FlowForm
from .navigator import Navigator
from .navigation_control import (
    NavigationOIDControls,
    NavigationHistoryControls,
    NavigationBar
)
from .script_line import ScriptLine


class CustomPageWidget(QtWidgets.QWidget):

    def __init__(self, host, session):
        super(CustomPageWidget, self).__init__(host)
        self.page = host.page
        self.session = session
        self.oid = None

    def set_oid(self, oid):
        self.oid = oid
        self.build()

    def build(self):
        '''
        Will be called once self.oid is set.
        Use self.session.cmds.Flow and self.oid to access
        the Object to show/edit.

        :return: None
        '''
        pass

    def on_touch_event(self, oid):
        '''
        Will be called to receive flow touch events.
        It's up to you do react on not, depending on the oid.

        :param oid: oid of the touched Object
        :return: None
        '''

    def die(self):
        '''
        Will be called when navigating away from the Object.
        :return: None
        '''
        pass

class MessagePageWidget(CustomPageWidget):

    def set_message(self, message):
        self._message = message

    def build(self):
        title = QtWidgets.QLabel(
            '<H1>Error:</H1>'
        )
        self.txt = QtWidgets.QTextEdit(self)
        self.txt.setReadOnly(True)
        self.txt.setText(self._message)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(title)
        self.layout().addWidget(self.txt)

class CustomPageHost(QtWidgets.QWidget):

    def __init__(self, page):
        super(CustomPageHost, self).__init__(page)
        self.page = page
        self._hosted = None
        self.setLayout(QtWidgets.QVBoxLayout())

    def has_host(self):
        return bool(self._hosted)

    def unhost(self):
        self.hide()
        if self._hosted is not None:
            self._hosted.die()

        # Clear all layout items (including spacing stuff)
        li = self.layout().takeAt(0)
        while li is not None:
            w = li.widget()
            if w is not None:
                w.setParent(None)
                w.deleteLater()
            li = self.layout().takeAt(0)

        if self._hosted is not None:
            self._hosted.deleteLater()
        self._hosted = None

    def _create_error_report(self, message):
        w = MessagePageWidget(self, self.page.session)
        w.set_message(message)
        return w

    def _create_hosted(self, qualified_type_name):
        # NB: str it because json tend to send unicodes here :/
        chunks = str(qualified_type_name).rsplit('.', 1)
        type_name = chunks.pop(-1)

        module_path = chunks[0]
        leaf_module_name = module_path.rsplit('.', 1)[-1]
        try:
            module = __import__(module_path, None, None, leaf_module_name, 0)
        except ImportError as err:
            return self._create_error_report(
                'Error while importing module for custom page {}:\n'
                '{}\nTrace:\n{}'.format(
                    qualified_type_name,
                    err,
                    traceback.format_exc()
                )
            )

        try:
            cls = getattr(module, type_name)
        except AttributeError as err:
            return self._create_error_report(
                'Error while creating custom page {}:\n'
                '{}\nTrace:\n{}'.format(
                    qualified_type_name,
                    err,
                    traceback.format_exc()
                )
            )

        w = cls(self, self.page.session)
        return w

    def host(self, oid, custom_page_str):
        if self._hosted is not None:
            self.unhost()
        self._hosted = self._create_hosted(custom_page_str)
        self.show()
        self._hosted.set_oid(oid)
        self.layout().addWidget(self._hosted)

    def dispatch_touch_event(self, oid):
        if self._hosted is not None:
            self._hosted.on_touch_event(oid)

class FlowPage(QtWidgets.QWidget):

    def __init__(self, parent, view, start_oid, root_oid):
        super(FlowPage, self).__init__(parent)

        self.view = view
        self.session = view.session

        self._navigator = Navigator(
            self.session, root_oid, start_oid
        )
        self._navigator.set_create_view_function(view.create_view)

        self.nav_bar = NavigationBar(self, self._navigator)
        self.nav_ctrl = self.nav_bar.nav_ctrl
        self.nav_oid = self.nav_bar.nav_oid

        self.custom_page_host = CustomPageHost(self)
        self.custom_page_host.hide()
        self.form = FlowForm(self, self)

        lo = QtWidgets.QVBoxLayout()
        lo.addWidget(self.nav_bar)
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(0)
        self.setLayout(lo)
        lo.addWidget(self.form, 100)
        lo.addWidget(self.custom_page_host, 100)

        self._navigator.add_on_current_changed(self.refresh)
        self._navigator.add_on_list_changed(self.nav_bar.nav_ctrl.update)

        self._source_view_id = None
        # self._dialogs = {}  # oid to dialg widget

        # FIXME: wtf is this ?!? why isn't this in NavigationBar class ?!?
        self.nav_bar.setAttribute(QtCore.Qt.WA_StyledBackground, True)

    def root_oid(self):
        return self._navigator.root_oid()
    
    def set_root_oid(self, oid):
        return self._navigator.apply_new_root_oid(oid)

    def current_oid(self):
        return self._navigator.current_oid()

    def set_source_view_id(self, source_view_id):
        '''
        Used by run_action to find affected vie by stuffs like 'goto' request
        '''
        self._source_view_id = source_view_id

    def history_oids(self):
        return self._navigator.history_oids()

    def set_history_oids(self, oids):
        return self._navigator.apply_history_oids(oids)

    def has_custom_page(self):
        return self.custom_page_host.has_host()

    def open(self, oid):
        self._navigator.goto(oid, new_view=True)

    def goto(self, oid):
        in_new_view = (
                QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier
        )
        if oid != self._navigator.current_oid() and not in_new_view:
            # This is requiered to fix a focus issue in some specific cases...
            self.setFocus()
        self._navigator.goto(oid, in_new_view)
        self.session.dispatch_event("focus_changed", view_id=self.view._view_id)

    def goto_connected(self, oid):
        in_new_view = (
                QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier
        )
        self._navigator.goto_parent_of(oid, in_new_view)

    def show_action_dialog(self, action_oid):
        if not self.session.cmds.Flow.action_needs_dialog(action_oid):
            self.run_action(action_oid, None)
            return

        # try:
        #     dialog = self._dialogs[action_oid]
        # except:
        #     pass
        # else:
        #     dialog.show()
        #     dialog.raise_()
        #     dialog.activateWindow()
        #     return

        form_config = self.form.config()
        ui = self.session.cmds.Flow.get_object_ui(action_oid)
        size = ui.get('dialog_size')
        min_size = ui.get('dialog_min_size')
        max_size = ui.get('dialog_max_size')

        if size is not None:
            form_config['dialog_size'] = size
        if min_size is not None:
            form_config['dialog_min_size'] = min_size
        if max_size is not None:
            form_config['dialog_max_size'] = max_size

        dialog = FlowDialogView(
            self.session,
            view_id=None,
            source_view_id=self.view.view_id(),
            parent_widget=self.view,
            start_oid=action_oid,
            root_oid=action_oid,
            form_config=form_config
        )

        # def on_dialog_close(result, page=self, oid=action_oid):
        #     # dialog = page._dialogs.pop(oid)
        #     dialog.forget_dialog_view()
        #     dialog.deleteLater()
        # dialog.finished.connect(on_dialog_close)

        # self._dialogs[action_oid] = dialog

        dialog.show()

    def run_action(self, action_oid, button):

        result = self.session.cmds.Flow.run_action(action_oid, button)
        result = result or {}

        if result.get('refresh', False):
            self.refresh()

        default_view_type = 'Flow'
        goto_oid = result.get('goto')
        if goto_oid is not None:
            target = result.get('goto_target', self._source_view_id)
            if target is None:
                target = self.view.view_id()
            view_type_name = result.get('goto_target_type', default_view_type)
            view = self.session.find_view(
                view_type_name=view_type_name,
                view_id=target
            )
            if view is None:
                new_view_id = None
                if target != '_NEW_':
                    new_view_id = target
                view = self.session.add_view(
                    view_type_name=view_type_name,
                    view_id=new_view_id
                )
            else:
                view.ensure_visible()
            view.goto_request(goto_oid)

        next_oid = result.get('next_action')
        if next_oid is not None:
            self.goto(next_oid)
            return

        if action_oid == self.current_oid() and result.get('close', True):
            self.view.on_action_result_close()

    def clear(self):
        self.form.clear()

    def refresh(self):
        oid = self.current_oid()
        view_title = self.session.cmds.Flow.get_source_display(oid)
        self.view.set_view_title(view_title)

        self.clear()

        self.nav_oid.update()
        self.nav_ctrl.update()

        ui = self.session.cmds.Flow.get_object_ui(oid)
        self.view.set_show_navigation_bar(ui.get('navigation_bar', True))
        custom_page = ui.get('custom_page')
        if custom_page:
            self.custom_page_host.host(oid, custom_page)
            self.form.hide()
        else:
            self.custom_page_host.unhost()
            self.form.show()
            self.form.build_roots(oid)
        
        # Update session layout autosave
        if (
            self.session.layout_autosave 
            and self.session.layout_load is False
            and (isinstance(self.view, FlowView) and self.view.dock_widget())
        ):
            self.session.store_layout_preset(
                self.session.get_views_state(main_geometry=True),
                autosave=True
            )

    def receive_event(self, event, data):
        # print('   --- FlowPage got event', event, data)
        if event == 'flow_touched':
            oid = data['oid']
            self.custom_page_host.dispatch_touch_event(oid)
            root = self.form.invisibleRootItem()
            for i in range(root.childCount()):
                item = root.child(i)
                item.on_touch_event(oid=oid)
            # for oid, dialog in six.iteritems(self._dialogs):
            #     page = dialog.flow_page
            #     page.receive_event(event, data)
        elif event == "focus_changed":
            view_id = data["view_id"]
            if view_id == self.view._view_id:
                self.setProperty("current", True)
                self.setAttribute(QtCore.Qt.WA_StyledBackground)
                self.form.style().polish(self.form)
                self.nav_bar.style().polish(self.nav_bar)
            else:
                self.setProperty("current", False)
                self.setAttribute(QtCore.Qt.WA_StyledBackground)
                self.form.style().polish(self.form)
                self.nav_bar.style().polish(self.nav_bar)

class FlowDialogView(DialogView):

    def __init__(self, session, view_id, parent_widget, source_view_id, start_oid, root_oid, form_config):
        super(FlowDialogView, self).__init__(session, view_id, parent_widget)

        self.setWindowTitle(root_oid)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.flow_page = FlowPage(self, self, start_oid, root_oid)
        self.flow_page.set_source_view_id(source_view_id)
        self.flow_page.form.configure(**form_config)

        self.layout().addWidget(self.flow_page)

        sg = QtWidgets.QSizeGrip(self)
        sg.setMinimumSize(5, 5)
        self.layout().addWidget(sg, 0, QtCore.Qt.AlignRight)

        self.installEventFilter(event_filters.EatEnterKeys())
        self.flow_page.form.viewport().installEventFilter(
            event_filters.MouseResizer(self)
        )

    def set_show_navigation_bar(self, b):
        self.flow_page.nav_bar.setVisible(b)

    def goto_request(self, oid):
        self.flow_page.goto(oid)

    def create_view(self, oid=None):
        # Needed as being the view of a FlowPage()
        pass

    def fitSize(self):
        form_config = self.flow_page.form.config()
        w, h = form_config.get('dialog_size', (-1, -1))
        min_w, min_h = form_config.get('dialog_min_size', (-1, -1))
        max_w, max_h = form_config.get('dialog_max_size', (-1, -1))

        if all(s < 0 for s in (w, h, min_w, min_h, max_w, max_h)):
            # No option or all options negative: revert to default behaviour
            cw, ch = self.flow_page.form.content_size()
            self.flow_page.form.setMinimumSize(cw, ch)
            self.flow_page.adjustSize()
            self.adjustSize()
        else:
            # Negative or zero size: resize to content
            if (w, h) <= (0, 0):
                (w, h) = self.flow_page.form.content_size()

            # Set min and max sizes only if they are valid
            if (0, 0) <= (min_w, min_h) and ((max_w, max_h) < (0, 0) or (min_w, min_h) <= (max_w, max_h)):
                self.setMinimumSize(min_w, min_h)
                w = max(w, min_w)
                h = max(h, min_h)
            
            if (0, 0) < (max_w, max_h) and ((min_w, min_h) < (0, 0) or (min_w, min_h) <= (max_w, max_h)):
                self.setMaximumSize(max_w, max_h)
                w = min(w, max_w)
                h = min(h, max_h)
        
        if (w, h) != (-1, -1):
            self.resize(w, h)

    def show(self):
        super(FlowDialogView, self).show()
        self.fitSize()
        self.moveToCenter()

    def receive_event(self, event, data):
        # print('   --- FlowDialogView got event', event, data)
        if event == 'flow_touched':
            self.flow_page.receive_event(event, data)

    def on_action_result_close(self):
        self.accept()


# class XX_FlowDialog(QtWidgets.QDialog):

#     def __init__(self, master_page, parent, view, start_oid, root_oid):
#         super(FlowDialog, self).__init__(parent)

#         self.setWindowTitle(root_oid)

#         self.setLayout(QtWidgets.QVBoxLayout())
#         self.layout().setContentsMargins(0, 0, 0, 0)

#         self.flow_page = FlowPage(self, view, start_oid, root_oid)
#         self.flow_page.set_master_page(master_page)
#         self.flow_page.form.configure(**master_page.form.config())

#         self.layout().addWidget(self.flow_page)

#         sg = QtWidgets.QSizeGrip(self)
#         sg.setMinimumSize(5,5)
#         self.layout().addWidget(sg, 0, QtCore.Qt.AlignRight)

#         self.installEventFilter(event_filters.EatEnterKeys())
#         self.flow_page.form.viewport().installEventFilter(
#             event_filters.MouseResizer(self)
#         )

#         self.setProperty("dialog", True)

#     def fitSize(self):
#         self.flow_page.form.fitSize()
#         self.flow_page.adjustSize()
#         self.adjustSize()

#     def show(self):
#         super(FlowDialog, self).show()
#         self.flow_page.refresh()
#         self.fitSize()

class FlowView(DockedView):

    @classmethod
    def view_type_name(cls):
        return 'Flow'

    def __init__(self, session, view_id=None, hidden=False, area=None, oid=None, root_oid=None):
        self._start_oid = oid
        self._root_oid = root_oid
        self.dev_menu = None
        self.script_line = None
        self.flow_page = None
        super(FlowView, self).__init__(session, view_id, hidden=hidden, area=area)

    def build_top(self, top_parent, top_layout, header_parent, header_layout):
        self.add_header_tool('*', '*', 'Duplicate View', self.create_view)

        self.view_menu.setTitle('Options')

        a = self.view_menu.addAction('Show Navigation Bar')
        a.setCheckable(True)
        a.setChecked(True)
        a.toggled.connect(self.set_show_navigation_bar)
        self._show_nav_bar_action = a

        a = self.view_menu.addAction('Show Hidden Relations')
        a.setCheckable(True)
        a.setChecked(False)
        a.toggled.connect(self.set_show_hidden_relations)

        a = self.view_menu.addAction('Show References')
        a.setCheckable(True)
        a.setChecked(False)
        a.toggled.connect(self.set_show_references_relations)

        self.view_menu.addAction('Create New View')
        self.view_menu.addSeparator()
        self.view_menu.addAction(
            'Activate DEV Tools',
            self._activate_dev_tools
        )

        self.script_line = ScriptLine(top_parent, self)
        self.script_line.hide()
        top_layout.addWidget(self.script_line, 100)

    def build_page(self, main_parent):
        self.flow_page = FlowPage(
            main_parent, self, self._start_oid, self._root_oid
        )

        lo = QtWidgets.QVBoxLayout()
        lo.setContentsMargins(0, 0, 0, 0)
        lo.addWidget(self.flow_page)
        self.flow_page.show()

        main_parent.setLayout(lo)
        self.flow_page.refresh()

    def _build(self, top_parent, top_layout, main_parent, header_parent, header_layout):
        self.build_top(top_parent, top_layout, header_parent, header_layout)
        self.build_page(main_parent)

    def get_view_state(self):
        self.setObjectName(self.view_id())
        self.dock_widget().setObjectName(self.view_id())

        return dict(
            oid=self.flow_page.current_oid(),
            root_oid=self._root_oid,
            history_oids=self.flow_page.history_oids(),
            form_config=self.flow_page.form.config()
        )

    def set_view_state(self, state):
        oid = state.get('oid', self.flow_page.current_oid())
        root_oid = state.get('root_oid', self._root_oid)
        history_oids = state.get('history_oids', [])
        form_config = state.get('form_config')

        if form_config is not None and self.flow_page.has_custom_page() is False:
            self.flow_page.form.configure(**form_config)

        self._start_oid = oid
        self._root_oid = root_oid
        self.flow_page.set_root_oid(root_oid)
        self.flow_page.set_history_oids(history_oids)

    def set_show_references_relations(self, b):
        self.flow_page.form.configure(show_references_relation=b)

    def set_show_navigation_bar(self, b):
        self.flow_page.nav_bar.setVisible(b)
        self._show_nav_bar_action.setChecked(b)

    def set_show_hidden_relations(self, b):
        self.flow_page.form.configure(show_hidden_relations=b)

    def set_show_protected_relations(self, b):
        self.flow_page.form.configure(show_protected_relations=b)

    def set_group_relations(self, b):
        self.flow_page.form.configure(group_relations=b)

    def toggle_script_line(self):
        self.script_line.setVisible(not self.script_line.isVisible())

    def reload_projects(self):
        self.session.cmds.Flow.reload_project(self.flow_page.current_oid())
        self.flow_page.refresh()

    def _activate_dev_tools(self):
        if self.dev_menu is not None:
            return
        self.dev_menu = self.menubar.addMenu('[DEV]')

        self.dev_menu.addAction('Toggle Script Line', self.toggle_script_line)

        a = self.view_menu.addAction('Group Relations')
        a.setCheckable(True)
        a.setChecked(True)
        a.toggled.connect(self.set_group_relations)

        a = self.dev_menu.addAction('Show Protected Relations')
        a.setCheckable(True)
        a.setChecked(False)
        a.toggled.connect(self.set_show_protected_relations)

        self.dev_menu.addSeparator()

        self.dev_menu.addAction('Reload Projects Definition', self.reload_projects)

        self.toggle_script_line()

    def create_view(self, oid=None):
        if oid is None:
            oid = self.flow_page.current_oid()
        self.duplicate_view(oid=oid)

    def receive_event(self, event, data):
        # print('   --- FlowView got event', event, data)
        if event == 'flow_touched':
            self.flow_page.receive_event(event, data)
        elif event == 'focus_changed':
            self.flow_page.receive_event(event, data)

    def on_action_result_close(self):
        '''
        This is called when an action show in self.flow_page is run and its result
        requests closing the containing view.
        This has meaning only for a DialogView so nothing is done here.
        '''
        pass

    def goto_request(self, oid):
        self.flow_page.goto(oid)


class FlowContextView(FlowView):

    @classmethod
    def get_context_oid(cls):
        return os.environ.get('KABARET_CONTEXT_OID')

    @classmethod
    def view_type_name(cls):
        return 'Context'

    def __init__(
        self, session, view_id=None, hidden=False, area=None,
    ):
        self._context_oid = self.get_context_oid()
        super(FlowContextView, self).__init__(
            session, view_id=view_id, hidden=hidden, area=area,
            oid=self._context_oid, root_oid=self._context_oid
        )

        self.set_view_title("Kabaret Context")

    def create_view(self, oid=None):
        if oid is None:
            oid = self.get_context_oid()
        self.duplicate_view(oid=oid)




import sys
import argparse

from kabaret.app.ui import gui

from .icons import gui as _  # register our icons.gui resourses
from .icons import flow as _  # register our icons.flow resourses


from kabaret.app.ui.gui.styles import Style
from qtpy import QtWidgets, QtGui

from kabaret.app.ui.gui.widgets.session_toolbar import SessionToolBar

# To test the custom home:
CUSTOM_HOME = True
CUSTOM_HOME = False
if CUSTOM_HOME:
    from kabaret.app.actors.flow import Flow
    from .flow.custom_home import MyHomeRoot

# In case you have it installed:
SCRIPT_VIEW = True
try:
    from kabaret.script_view import script_view
except ImportError:
    SCRIPT_VIEW = False

class MyStyle(Style):
    '''
    This is to show how to add a style to the app, and make it the default one.
    Beware that a style needs more than this to be readable (tabbed dock window and
    splitters need some luvin'!)

    '''

    def __init__(self):
        super(MyStyle, self).__init__('MyStyle')

    def apply(self, widget):

        app = QtWidgets.QApplication.instance()

        if widget is app:
            # setup the palette
            palette = QtWidgets.QApplication.palette()

            # A color to indicate a selected item or the current item. By default, the highlight color is Qt.darkBlue.
            palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(80, 0, 0))

            QtWidgets.QApplication.setPalette(palette)

# Uncomment this to activate MyStyle
# MyStyle() # install my style (the last installed is set as default)


class MyStudioGUISession(gui.KabaretStandaloneGUISession):

    def register_view_types(self):
        super(MyStudioGUISession, self).register_view_types()

        type_name = self.register_view_type(SessionToolBar)
        self.add_view(type_name)

        if SCRIPT_VIEW:
            type_name = self.register_view_type(script_view.ScriptView)
            self.add_view(type_name)

    def _create_actors(self):
        '''
        Instanciate the session actors.
        Subclasses can override this to install customs actors or
        replace default ones.
        '''
        if CUSTOM_HOME:
            Flow(self, CustomHomeRootType=MyHomeRoot)
        else:
            return super(MyStudioGUISession, self)._create_actors()


def process_remaining_args(args):
    parser = argparse.ArgumentParser(
        description='Kabaret Session Arguments'
    )
    parser.add_argument(
        '--layout-mgr', default=False, action='store_true', dest='layout_mgr', help='Enable Layout Manager'
    )
    parser.add_argument(
        '--no-layout-mgr', action='store_false', dest='layout_mgr', help='Disable Layout Manager'
    )
    parser.add_argument(
        '--layout-autosave', default=False, action='store_true', dest='layout_autosave', help='Use Layout Autosave'
    )
    parser.add_argument(
        '--no-layout-autosave', action='store_false', dest='layout_autosave', help='Disable Layout Autosave'
    )
    parser.add_argument(
        '--layout-savepath', default=None, dest='layout_savepath', help='Specify Layout Saves Path'
    )
    values, _ = parser.parse_known_args(args)

    return (
        values.layout_mgr,
        values.layout_autosave,
        values.layout_savepath
    )


if __name__ == '__main__':
    argv = sys.argv[1:]  # get ride of first args wich is script filename
    (
        session_name,
        host, port, cluster_name,
        db_index, password, debug,
        remaining_args
    ) = MyStudioGUISession.parse_command_line_args(argv)

    layout_mgr, layout_autosave, layout_savepath = process_remaining_args(remaining_args)

    session = MyStudioGUISession(
        session_name=session_name, debug=debug,
        layout_mgr=layout_mgr, layout_autosave=layout_autosave, layout_savepath=layout_savepath
    )
    session.cmds.Cluster.connect(host, port, cluster_name, db_index, password)

    session.start()
    session.close()





from . import styles                        # init the style manager
from .styles import neonative, dark, gray

neonative.NeonativeStyle()  # registers the style in the StyleManager
gray.GrayStyle()            # # registers the style in the StyleManager
dark.DarkStyle()            # registers the style in the StyleManager, last one is default one.


from .standalone import KabaretStandaloneGUISession
from .embedded import KabaretEmbeddedGUISession



'''

    kabaret.flow.object

    Defines the Object class, the base for all object in a flow, and
    the Root class used to hold a flow graph.

'''
import logging

import six

from .exceptions import WIPException
from .exceptions import MissingRelationError, MissingChildError

from .value_store import MemoryValueStore


class _Manager(object):

    def __init__(self, object, parent, name, value_store):
        super(_Manager, self).__init__()
        self.object = object
        self.parent = parent
        self.name = name
        self.value_store = value_store
        self.destroyed = False  # TODO: use this to stop behavior after desctuction

        if self.parent is not None:
            self.parent._mng.set_child(self, object)

        self.children = {}

#---OLD
    def create_manager(self, object, parent, name):
        raise WIPException('Is this used anywhere ?!?')
        return self.__class__(object, parent, name)

    def root(self):
        if self.parent is None:
            return self.object
        return self.parent._mng.root()

#---OLD
    def iter_parents(self):
        raise WIPException('Is this used anywhere ?!?')
        curr = self
        while curr is not None:
            parent = curr.parent
            if parent is None:
                raise StopIteration()
            yield parent
            curr = parent._mng

    def oid(self):
        # TODO: since we cant rename an item, and we use this quite a lot:
        # shouldn't we bake it ?
        # (it may help for Map.rows())
        if self.parent is None:
            return self.name
        return '%s/%s' % (self.parent._mng.oid(), self.name)

    def destroy(self):
        logging.getLogger('kabaret.flow').debug('----> DESTROYING ' + self.oid() + '. Should we notify refs ?')
        self.destroyed = True
        self.parent = None

    def set_child(self, mng, object):
        try:
            self.children[mng.name]
        except:
            self.children[mng.name] = object
        else:
            raise Exception('This child already exists !!!')

    def relations(self):
        return self.object._relations

#---OLD
    def drop_child(self, name):
        raise WIPException('This has been renamed destroy_child')

    def destroy_child(self, name):
        try:
            old = self.children[name]
        except KeyError:
            raise MissingChildError(self.oid(), name)
        else:
            old._mng.destroy()
        del self.children[name]

#---OLD
    def children_names(self):
        # relations must be used to find children names.
        raise WIPException('This should not be useful')
        try:
            return self.object.children_names()
        except AttributeError:
            return self.children.keys()

#---OLD
    def children_oids(self):
        raise WIPException('This should not be useful')
        oid = self.oid()
        return ['%s/%s' % (oid, n) for n in self.children_names()]

#---OLD
    def object_type_name(self):
        raise WIPException('Is this used anywhere ?!?')
        return self.object.__class__.__name__

    def get_qualified_type_name(self, TYPE):
        return '%s.%s' % (TYPE.__module__, TYPE.__name__)

    def object_qualified_type_name(self):
        return self.get_qualified_type_name(self.object.__class__)

    def qualified_type_name_to_type(self, qualified_type_name):
        # NB: str it because json tend to send unicodes here :/
        chunks = str(qualified_type_name).rsplit('.', 1)
        type_name = chunks.pop(-1)

        module_path = chunks[0]
        leaf_module_name = module_path.rsplit('.', 1)[-1]
        module = __import__(module_path, None, None, leaf_module_name, 0)

        try:
            return getattr(module, type_name)
        except AttributeError:
            raise ImportError("Nothing named %s in module %s" %
                              (type_name, module_path))

#---OLD
    def __len__(self):
        raise WIPException('This should not be useful')
        return len(self.children)

#---OLD
    def __iter__(self):
        raise WIPException('This should not be useful')
        for name, object in self.children.iteritems():
            yield name, object

#---OLD
    def __contains__(self, object_name):
        raise WIPException('This should not be useful')
        return object_name in self.children

    def has_related(self, relation_name):
        return relation_name in [r.name for r in self.object._relations]
    
    def get_relation(self, name):
        try:
            return next((r for r in self.object._relations if r.name == name))
        except StopIteration:
            raise AttributeError("Could not find relation '%s' under '%s'" %
                                (name, self.oid()))

#---OLD
    def get_child(self, name):
        raise WIPException('This has been renamed to get_existing_child(name)')

    def get_existing_child(self, name):
        '''
        Return the child object named 'name'.
        This DOES NOT involve Relation instance and can only return already
        existing children.

        If no such child exists, a MissingChildError is raised.
        '''
        try:
            return self.children[name]
        except KeyError:
            raise MissingChildError(self.oid(), name)

#---OLD
    def get_typed_children(self, class_or_type_or_tuple):
        raise WIPException('Is this used anywhere ?!?')
        return [
            relation.get_related(self.object)
            for relation in self.object._relations
            if issubclass(relation.related_type, class_or_type_or_tuple)
        ]

    def get_object(self, oid):
        '''
        Returns the object with the given oid.
        If oid is relative (does not start with /), returns a (grand) child.

        The oid may contain . and .. has in posix paths
        (do we really use that ? o.O)

        This will evaluate the _Relation used by self.object and potential
        override of __getattr__.
        '''
        if oid is None:
            return self.object

        # remove consecutive slashes (//) to avoid requesting root in
        # the middle of the search
        while '//' in oid:
            oid = oid.replace('//', '/')

        cuts = oid.split('/', 1)
        try:
            remain = cuts[1]
        except IndexError:
            remain = None

        next = cuts[0]

        if not next:
            root = self.root()
            return root._mng.get_object(remain)

        if next == '.':
            return self.get_object(remain)

        if next == '..':
            return self.parent._mng.get_object(remain)

        if self.has_related(next):
            # we must us getattr to have the relation do the job:
            child = getattr(self.object, next)
            if child is None:
                # This may happen with erroneous `Relative` oid
                return None
        else:
            # maybe it's in a map
            try:
                self.object.get_mapped
            except AttributeError:
                raise MissingRelationError(self.oid(), next)
            else:
                child = self.object.get_mapped(next)

        return child._mng.get_object(remain)

#---OLD
    def set_default_value(self, value):
        raise WIPException('! this should be in Value !')
        self.default_value = value

#---OLD
    def compute_value(self):
        raise WIPException('! this should be in ComputedValue !')
        if not self.is_dirty():
            return
        self.parent.compute_child_value(self.object)

    def add_ref(self, object):
        oid = object.oid()
        refs = set(self.refs())
        if oid in refs:
            # Avoid doubles and extra refresh/events:
            return
        refs.add(object.oid())
        refs = sorted(refs)
        self.value_store.set(self.oid() + '.refs', refs)

    def remove_ref(self, object):
        # use set to cleanup duplicates (was possible in previous code)
        refs = set(self.refs())
        try:
            refs.remove(object.oid())
        except KeyError:
            # This should not happen.
            logging.getLogger('kabaret.flow').warning('Removing ref %r from %r: ref not in list!')
            return
        refs = sorted(refs)
        self.value_store.set(self.oid() + '.refs', refs)

    def refs(self):
        try:
            return self.value_store.get(self.oid() + '.refs')
        except KeyError:
            return []

    def cleanup_refs(self):
        """
        check registered refs and remove those which are not relevant anymore (deleted object or not connected anymore)
        """
        try:
            refs = self.value_store.get(self.oid() + '.refs')
        except KeyError:
            return

        valid_refs = set()
        for ref_oid in refs:
            try:
                ref = self.get_object(ref_oid)
            except Exception as e:
                logging.getLogger('kabaret').warning("Cannot find {}: {}".format(ref_oid, e))
            else:
                if ref.get() == self.object:
                    valid_refs.add(ref_oid)

        self.value_store.set(self.oid() + '.refs', list(valid_refs))

    def ref_objects(self):
        return [self.get_object(ref_oid) for ref_oid in self.refs()]

    def get_value(self, default):
        try:
            return self.value_store.get(self.oid())
        except KeyError:
            return default

    def del_value(self):
        self.value_store.delete(self.oid())
        self._on_value_changed()

    def set_value(self, value):
        self.value_store.set(self.oid(), value)
        self._on_value_changed()

#--- OLD
    # def update_dict_value(self, **new_values):
    #     self.value_store.update(self.oid(), **new_values)
    #     self._on_value_changed()

    def incr_value(self, by=1):
        self.value_store.incr(self.oid(), by)
        self._on_value_changed()

    def decr_value(self, by=1):
        self.value_store.decr(self.oid(), by)
        self._on_value_changed()

    # --- OSS

    def oss_get(self):
        return self.value_store.oss_get(self.oid())

    def oss_get_range(self, first, last):
        return self.value_store.oss_get_range(self.oid(), first, last)

    def oss_has(self, member):
        return self.value_store.oss_has(self.oid(), member)

    def oss_add(self, member, score):
        self.value_store.oss_add(self.oid(), member, score)
        self._on_value_changed()

    def oss_remove(self, member):
        self.value_store.oss_remove(self.oid(), member)
        self._on_value_changed()

    def oss_len(self):
        return self.value_store.oss_len(self.oid())

    def oss_get_score(self, member, score):
        return self.value_store.oss_get_score(self.oid(), member)

    def oss_set_score(self, member, score):
        self.value_store.oss_set_score(self.oid(), member, score)
        self._on_value_changed()

    # --- HASH

    def get_hash_value(self):
        return self.value_store.get(self.oid())

    def hash_get_key(self, key):
        return self.value_store.hash_get_key(self.oid(), key)

    def hash_has_key(self, key):
        return self.value_store.hash_has_key(self.oid(), key)

    def del_hash_key(self, key):
        self.value_store.del_hash_key(self.oid(), key)
        self._on_value_changed()

    def get_hash(self):
        return self.value_store.get_hash(self.oid())

    def get_hash_as_dict(self):
        return self.value_store.get_hash_as_dict(self.oid())

    def get_hash_keys(self):
        return self.value_store.get_hash_keys(self.oid())

    def get_hash_len(self):
        return self.value_store.get_hash_len(self.oid())

    def update_hash(self, mapping):
        self.value_store.update_hash(self.oid(), mapping)
        self._on_value_changed()

    def set_hash(self, mapping):
        self.value_store.set_hash(self.oid(), mapping)
        self._on_value_changed()

    def set_hash_key(self, key, value):
        self.value_store.set_hash_key(self.oid(), key, value)
        self._on_value_changed()

    def _on_value_changed(self):
        # use self.object.touch() and not self.touch()
        # because object.touch() may have been overridden to implement
        # touch propagation.
        self.object.touch()

    def touch(self):
        self.root().object_touched(self.object)

    def pformat(self, indent=0):
        from .values import Value, Ref

        if isinstance(self.object, Ref):
            target = self.object.get()
            if target is not None:
                infos = "%s<%s>(%s)" % (
                    self.object.__class__.__name__,
                    target.__class__.__name__,
                    target._mng.oid()
                )
            else:
                infos = "%s<%s>" % (
                    self.object.__class__.__name__, target.__class__.__name__)
        elif isinstance(self.object, Value):
            infos = "%s(%s)" % (
                self.object.__class__.__name__, repr(self.object.get())[:50])
        else:
            infos = self.object.__class__.__name__
        head = "%s%s = %s" % (
            indent * 0 * '  ',
            self.oid(),
            infos,
        )
        i = indent + 1
        body = '\n'.join([
            o._mng.pformat(i) for n, o in sorted(self.children.items())
        ])
        return head + (body and '\n' + body or '')
    
    def on_object_init(self):
        pass


class _Relation(object):
    '''
    This is the base class of all relations.
    You must use one of the subclasses.

    Relations are descriptors managing the instantiation and the access
    to object related to the relation owner.

    You can configure the relation behavior using the :any:`ui()` and :any:`editor()`
    methods.
    '''
    _RELATION_TYPE_NAME = None  # see cls.relation_type_name()

    MAX_INDEX = 100  # maximum automatic relation index.

    _next_relation_id = 0

    _GROUP_CONTEXT = []

    @classmethod
    def _generate_relation_id(cls):
        # tricky stuff to order relations in declaration order [part 1]
        # /!\ be sure to use _Relation class (not cls)
        _Relation._next_relation_id += 1
        return cls._next_relation_id

    @classmethod
    def relation_type_name(cls):
        '''
        Used by clients to classify relations.
        Returns cls._RELATION_TYPE_NAME. Subclass can set this
        class attribute to alter the string returned here.
        '''
        return cls._RELATION_TYPE_NAME

    @classmethod
    def get_default_group(cls):
        if not cls._GROUP_CONTEXT:
            return None
        return '.'.join(cls._GROUP_CONTEXT)

    def __init__(self, related_type):
        super(_Relation, self).__init__()
        self.id = self.__class__._generate_relation_id()
        self.index = 0

        self.group = None

        self.name = None
        self.related_type = related_type
        self.used = False

        self._ui = {
            # missing key means use default: 'editor_type': None,
            'icon': None,
            'editor_type': None,
            'editable': False,
            'label': None,
            'group': self.__class__.get_default_group(),
            'hidden': False,
            'tooltip': None,
        }

    def ui(
        self,
        icon=None,
        editor=None,
        editable=None,
        label=None,
        group=None,
        hidden=None,
        tooltip=None,
        expanded=None,
        expandable=None,
        **editor_options
    ):
        '''
        Configure the relation GUI.
        This method returns self so that you can chain
        it in assigment:
        ::
            meh = MyRelation('example').ui(label='Amazing')

        '''
        if icon is not None:
            self._ui['icon'] = icon
        if editor is not None:
            self._ui['editor_type'] = editor
        if editable is not None:
            self._ui['editable'] = editable
        if label is not None:
            self._ui['label'] = label
        if group is not None:
            self._ui['group'] = group
        if hidden is not None:
            self._ui['hidden'] = hidden and True or False
        if tooltip is not None:
            self._ui['tooltip'] = tooltip
        if expanded is not None:
            self._ui['expanded'] = bool(expanded)
        if expandable is not None:
            self._ui['expandable'] = bool(expandable)
        if editor_options:
            self._ui.update(editor_options)
        return self

    def is_hidden(self, of=None):
        ui = self.get_ui(of)
        return ui.get('hidden', False)

    def get_ui(self, of=None):
        '''

        :param of: the parent of the related object. If not None, the related
        Object._fill_ui(ui) is used instead of the default value set on the
        relation.
        :return: dict with keys icon, editor_type, editable, label, hidden, tooltip
        '''
        ui = self._ui.copy()
        if not ui.get('icon') and self.related_type is not None:
            ui['icon'] = self.related_type.ICON
        if of is not None:
            related = self.get_related(of)
            related._fill_ui(ui)
        return ui

    def set_related_type(self, related_type):
        if self.used:
            raise RuntimeError(
                'Too late to change the related type: '
                'some related object have already been created.'
            )
        self.related_type = related_type

    def _create_object(self, parent):
        raise NotImplementedError()

    def get_related(self, of):
        try:
            return of._mng.get_existing_child(self.name)
        except MissingChildError:
            pass  # let's create it

        related = self._create_object(of)
        self.used = True
        return related

    def __get__(self, o, t=None):
        '''
        Returns the child object referred to by this relation
        when it is accessed from an object instance. Otherwise
        (relation accessed from the class), returns the relation
        itself.
        '''
        if o is None:
            return self
        return self.get_related(o)


class _RelationGroupContext:

    def __init__(self, name):
        self.name = name

    def __enter__(self):
        _Relation._GROUP_CONTEXT.append(self.name)
        return self

    def __exit__(self, type, value, tb):
        _Relation._GROUP_CONTEXT.pop()


def group(name):
    return _RelationGroupContext(name)


class ObjectType(type):

    def __new__(cls, class_name, bases, class_dict):
        # tricky stuff to order relations in declaration order [part 2]
        default_relation_group = class_dict.get('LABEL', class_name)
        if bases != (object,):  # if it's not our most base class
            if 'oid' in class_dict:
                raise ValueError(
                    'Cannot create ObjectType %r with %r named "oid" (reserved name)' % (
                        class_name, class_dict['oid']
                    )
                )
            if 'name' in class_dict:
                raise ValueError('Cannot create ObjectType %r with %r named "name" (reserved name)' % (
                    class_name, class_dict['name']))
        else:
            default_relation_group = None

        relations = []
        relation_index = 0
        for base in bases:
            try:
                base_relations = base._relations
            except AttributeError:
                pass
            else:
                relations.extend(base_relations)
                relation_indexes = [
                    r.index for r in base_relations if r.index > 0 and r.index < _Relation.MAX_INDEX]
                if relation_indexes:
                    relation_index = max(relation_index, max(relation_indexes))

        _relations = sorted(
            ((key, o) for (key, o) in class_dict.items() if isinstance(o, _Relation)),
            key=lambda x: x[1].id
        )

        for n, o in _relations:
            if isinstance(o, _Relation):
                o.name = n
                if not o.index:
                    relation_index = relation_index + 1
                    o.index = relation_index
                if o.group is None:
                    o.group = default_relation_group
                if o.index > _Relation.MAX_INDEX:
                    raise Exception('Relation index for %r of ObjectType %r exceeded the max of %s.' %
                                    (n, class_name, _Relation.MAX_INDEX))
                relations.append(o)

        # remove overridden relations by using only the last ones with same name:
        relations = dict([(r.name, r) for r in relations])

        def relation_sort_key(r):
            return r.index

        relations = sorted(
            six.viewvalues(relations),  # py 2+3 version of relations.values()
            key=relation_sort_key
        )

        class_dict['_ref_type'] = None
        class_dict['_relations'] = relations
        return super(ObjectType, cls).__new__(cls, class_name, bases, class_dict)


class Object(six.with_metaclass(ObjectType, object)):

    _MANAGER_TYPE = None
    _PROPAGATE_MANAGER_TYPE = None

    @classmethod
    def _create_value_store(cls, parent, name):
        '''
        Subclasses may override this class method to return
        the value store they want to use.
        Default is to return None, which will lead to using the parent's value_store
        '''
        return None

    ICON = 'object'

    @classmethod
    def ref_type(cls):
        if cls._ref_type is not None:
            return cls._ref_type

        from .values import Ref

        class MyTypeRef(Ref):
            SOURCE_TYPE = cls
        MyTypeRef.__name__ = cls.__name__ + 'Ref'
        cls._ref_type = MyTypeRef
        return cls._ref_type

    @classmethod
    def get_source_display(cls, oid):
        '''
        Returns the text to display when showing a Ref pointing to the object of type cls with
        the id oid.

        This is used by Connection relations' GUI.
        '''
        return oid
    
    @classmethod
    def _resolve_manager_type(cls, parent):
        manager_type = None

        if parent is None:
            if cls._MANAGER_TYPE is None:
                manager_type = _Manager
            elif not issubclass(cls._MANAGER_TYPE, _Manager):
                raise TypeError(
                    'Invalid manager type %r (it must be _Manager or a subclass).' % cls._MANAGER_TYPE
                )
            else:
                manager_type = cls._MANAGER_TYPE
        else:
            if cls._MANAGER_TYPE is None:
                if parent._resolve_propagate_manager_type():
                    manager_type = parent._mng.__class__
                else:
                    manager_type = _Manager
            elif not issubclass(cls._MANAGER_TYPE, _Manager):
                raise TypeError(
                    'Invalid manager type %r (it must be _Manager or a subclass).' % cls._MANAGER_TYPE
                )
            else:
                manager_type = cls._MANAGER_TYPE
        
        return manager_type

    def _resolve_propagate_manager_type(self):
        propagate = self.__class__._PROPAGATE_MANAGER_TYPE
        parent = self._mng.parent
        
        # Get first propagation flag defined in parents
        while propagate is None and parent is not None:
            propagate = parent.__class__._PROPAGATE_MANAGER_TYPE
            parent = parent._mng.parent
        
        # Handle project root propagation flag set to None
        propagate = (propagate is not None) and propagate
        
        return propagate

    def __new__(cls, parent, name, **kwargs):
        if parent is None:
            # we are creating a root, the name must be none and kwargs must have a value_store key:
            if name != '':
                raise TypeError(
                    'Cannot create root (object parent == None) with a name that is not empty string.')
            if 'value_store' not in kwargs:
                raise TypeError(
                    'Cannot create root (object parent == None) without "value_store" keyword argument.')
            value_store = kwargs['value_store']
            manager_type = cls._resolve_manager_type(parent)
        else:
            # we are creating an Object, name must not be empty or None and kwargs must be empty:
            if parent is None:
                raise TypeError(
                    'Cannot create an object without parent (Use a Root() as parent if needed).')
            if not isinstance(parent, Object):
                raise TypeError(
                    'parent must be an Object, not a %r' % (type(parent),))
            if not name:
                raise TypeError('Cannot create an object without name.')
            if kwargs:
                raise TypeError(
                    'Cannot create object with more arguments than "parent" and "name".')

            # name must also be a valid python attribute name:
            if '.' in name:
                raise TypeError(
                    'Invalid object name %r (it must be a valid attribute name).' % (name,))
            try:
                exec(name + '=None') in {}
            except Exception:
                raise TypeError(
                    'Invalid object name %r (it must be a valid attribute name).' % (name,))

            manager_type = cls._resolve_manager_type(parent)

            # if not overriden, the value store is the the parent's one
            value_store = cls._create_value_store(parent, name)
            if value_store is None:
                value_store = parent._mng.value_store

        instance = super(Object, cls).__new__(cls)
        instance._mng = manager_type(instance, parent, name, value_store)

        return instance

    def __init__(self, parent, name):
        super(Object, self).__init__()
        self._mng.on_object_init()

    def name(self):
        return self._mng.name

    def oid(self):
        return self._mng.oid()

    def root(self):
        return self._mng.root()

    def touch(self):
        '''
        Force notification that the object has changed.
        '''
        self._mng.touch()

    def child_value_changed(self, child_value):
        '''
        Called when a watched child Value has changed.
        See relations.Param.watched() and values.Value.watch()
        '''
        raise NotImplementedError(
            'missing child_value_changed(child_value) in %s' % (self.oid()))

    def compute_child_value(self, child_value):
        '''
        Called when a ComputeValue child need its value to be computed.

        This happens when a computed value was touched before someone
        asks for its value.

        You must set the result of the computation to the child_value.
        '''
        raise NotImplementedError(
            'missing compute_child_value(child_value) in %s (child name: %r)' % (self.oid(), child_value.name()))

    def _fill_ui(self, ui):
        '''
        Override this to dynamically alter the ui option set on the
        relation holding this Object:
            icon, editor_type, editable, label, hidden, tooltip

        Default is to keep all default value.

        :param ui: the ui info set on the relation holding this Object.
        :return: None
        '''
        pass

#---OLD #FIXME: replace this by a _summary child or a Summary relation...
    def summary(self):
        return None

    def log(self):
        logging.getLogger('kabaret.flow').info(self._mng.pformat())


class SessionObject(Object):
    '''
    The SessionObject overrides its default value store
    with a MemoryValueStore() (all value die when the session ends)

    As the default value store is inherited by parent Objects, the
    whole branch under this one will also be "in memory only".
    '''

    @classmethod
    def _create_value_store(cls, parent, name):
        return MemoryValueStore()


class Root(Object):

    def __new__(cls, value_store):
        return super(Root, cls).__new__(cls, None, '', value_store=value_store)

    def __init__(self, value_store):
        super(Root, self).__init__(None, '')
        self._object_touched_handler = []

    def get_mapped(self, name):
        '''
        Implementing this is needed to have Ref's working with
        source_oid specified as absolute.
        '''
        raise NotImplementedError()

    def add_object_touched_handler(self, handler):
        self._object_touched_handler.append(handler)

    def remove_object_touched_handler(self, handler):
        self._object_touched_handler.remove(handler)

    def object_touched(self, object):
        '''
        Called when any object under this root was touched.
        '''
        for handler in self._object_touched_handler:
            handler(object)


class ThumbnailInfo(Object):
    '''
    This object defines the interface needed to provide "object image view" in
    the GUI.

    You define an object's thumbnail by defining its get_thumbnail_object()
    method and returning a related ThumbnailInfo instance:
        class MyObject(Object):

            _thumbnail = Child(MyThumbnailIn)
    '''

    def is_resource(self):
        '''
        Must return True if your thumbnail source is a resource file.
        If this returns True, does methods need to be implemented:
            get_default_heigth()
            get_label()
            get_resource()
        '''
        raise NotImplementedError()

    def get_resource(self):
        '''
        If is_resource() returns True, this must return a 2d string:
            resource_folder_name, resource_name
        '''
        raise NotImplementedError()

    def is_image(self):
        '''
        Must return True if your thumbnail source is a single image.
        If this returns True, does methods need to be implemented:
            get_default_heigth()
            get_label()
            get_path()
        '''
        raise NotImplementedError()

    def is_sequence(self):
        '''
        Must return True if your thumbnail source is a sequence of images.
        If this returns True, does methods need to be implemented:
            get_default_heigth()
            get_label()
            get_path()
            get_first_and_last()
        '''
        raise NotImplementedError()

    def get_label(self):
        '''
        Returns the label to display on the thumbnail.
        A value of None will use the basename of the thumbnail source file.
        '''
        return None

    def get_path(self):
        '''
        Returns the path of the thumbnail source.
        If is_sequence() returns True, the path must contain a formater for the
        frame number:
            /path/to/my/images.%04d.jpg

        Supported images types are roughtly the same as a standard webbrowser.
        (I.e: no EXP support here.)
        '''
        raise NotImplementedError()

    def get_default_height(self):
        '''
        Returns the default height of the thumbnail when first show in GUI.
        '''
        return 50

    def get_first_last_fps(self):
        '''
        If is_sequence() returns True, this must return the index of the
        first and the last frames, and the frame per second rate
        '''
        raise NotImplementedError()



'''


'''


class Navigator(object):

    def __init__(self, session, root_oid=None, current_oid=None):
        super(Navigator, self).__init__()

        self.session = session

        self._root_oid = root_oid
        self._current_oid = current_oid or root_oid
        self._prev_list = []
        self._next_list = []

        self._on_current_changed = []  # list of callables
        #---OLD self._on_current_corrected = [] # list of callbles

        self._on_list_changed = []  # list of callables

        self._create_view_function = None

        self.refresh()

    def add_on_current_changed(self, callable):
        '''callable must accept no args'''
        self._on_current_changed.append(callable)

    def remove_on_current_changed(self, callable):
        self._on_current_changed.remove(callable)

#---OLD
    # def add_on_current_corrected(self, callable):
    #     self._on_current_corrected.append(callable)

    # def remove_on_current_corrected(self, callable):
    #     self._on_current_corrected.remove(callable)

    # def correct_current_oid(self, oid):
    #     self._current_oid = oid
    #     for callable in self._on_current_corrected:
    #         callable(self._current_oid)

    def add_on_list_changed(self, callable):
        '''callable must accept no args'''
        self._on_list_changed.append(callable)

    def remove_on_list_changed(self, callable):
        self._on_list_changed.remove(callable)

    def set_create_view_function(self, f):
        '''
        The callable f must accept one argument:
                the oid to load on the new view.
        '''
        self._create_view_function = f

    def create_view(self, oid):
        if self._create_view_function is None:
            raise Exception(
                'Could not create a new view, set_create_view_function() '
                'never called :/'
            )
        self._create_view_function(oid)

    def refresh(self):
        ''' -> goto current oid'''
        self.goto(self.current_oid())

    def root_oid(self):
        return self._root_oid

    def apply_new_root_oid(self, oid):
        self._root_oid = oid

    def current_oid(self):
        return self._current_oid

    def _apply_new_current_oid(self, oid):
        self._current_oid = self.session.cmds.Flow.resolve_path(oid)
        for callable in self._on_current_changed:
            callable()

    def goto(self, oid, new_view=False):
        if new_view:
            self.create_view(oid)
            return

        self._next_list = []
        if self._current_oid and oid != self._current_oid:
            self._prev_list.append(self._current_oid)
        self._apply_new_current_oid(oid)

    def history_oids(self):
        return self._prev_list

    def apply_history_oids(self, prev_list):
        self._prev_list = prev_list
        for callable in self._on_list_changed:
            callable()

    def has_prev(self):
        return self._prev_list and True or False

    def goto_prev(self):
        if not self._prev_list:
            return
        self._next_list.append(self._current_oid)
        oid = self._prev_list.pop(-1)
        self._apply_new_current_oid(oid)

    def has_next(self):
        return self._next_list and True or False

    def goto_next(self):
        if not self._next_list:
            return
        self._prev_list.append(self._current_oid)
        oid = self._next_list.pop(-1)
        self._apply_new_current_oid(oid)

    def has_parent(self):
        if self.root_oid():
            return self._current_oid != self.root_oid()
        return '/' in self._current_oid  # .lstrip('/')

    def goto_parent(self, new_view=False):
        self.goto_parent_of(self._current_oid, new_view)

    def goto_parent_of(self, oid, new_view=False):
        skip_maps = True
        parent_oid = self.session.cmds.Flow.get_parent_oid(oid, skip_maps)
        self.goto(parent_oid, new_view)

    def goto_related(self, relation_name, resolve_ref=True, new_view=False):
        related_oid = self.session.cmds.Flow.get_related_oid(
            self._current_oid, relation_name, resolve_ref
        )
        self.goto(related_oid, new_view)

    def goto_root(self):
        self.goto(self.root_oid())

    def split_current_oid(self):
        skip_maps = True
        label_to_oid = self.session.cmds.Flow.split_oid(
            self._current_oid, skip_maps, self._root_oid
        )
        return label_to_oid

    def get_navigable_oids(self, oid, send_to):
        return self.session.cmds.Flow.get_navigable_oids(oid)



from kabaret.app import plugin

from .flow import Flow
from .cluster import Cluster

@plugin(trylast=True)
def install_actors(session):
    """
    Installs the default implementations of mandatory Actors (Cluster & Flow)
    if no such Actors are already installed.
    """

    installed_actor_names = session.get_actor_names()

    if "Cluster" not in installed_actor_names:
        # No Cluster Actor installed, install default one:
        Cluster(session)

    if "Flow" not in installed_actor_names:
        # No Flow Actor installed, install default one:
        Flow(session)
        
    



'''

    kabaret.flow.values

    Defines the basic value types:
        Value:          Stores a python object. Provides set() and get() methods
                        The Value is the base for all other values types.
        
        SessionValue:   Store a python object, but does not save it to the value store.
                        This means the value get reset to its default for each session,
                        and that it is not shared with other sessions.

        IntValue
        FloatValue
        StringValue
        DictValue:     Subclasses of Value each with their suitable methods (see sources...)

        OrderedStringSetValue:      
                        Stores an ordered by score set of strings.

        HashValue:      Stores a 2D list of (str, str)
                        This should be prefered over DictValue if:
                            - Your keys and values are strings.
                            - You need direct key access has/get/set.

        ComputedValue:  Stores a python object generated by the parent's compute_child_value(value) method.
                        The parent Object must implement this method and call value.set(something).
                        If set_cached(True) is called, the computation is not requested unless touch() is
                        called.
                        If store_value(True) is called, the computed value is also stored in the
                        value store (may be usefull if other applications read it...).

        ChoiceValue:    Stores a value which is one of a predifined possible values.

        MultiChoiceValue: Stores a list where each item is one of a predifined possible values.

        Ref:            References an Object.
'''
from .exceptions import WIPException
from .exceptions import MissingChildError, MissingRelationError, RefSourceError, RefSourceTypeError
from .object import Object


class Value(Object):

    ICON = 'value'
    DEFAULT_EDITOR = None

    def __init__(self, parent, name):
        super(Value, self).__init__(parent, name)
        self._default_value = None
        self._watched = False

    def set_watched(self, b):
        self._watched = bool(b)

    def notify(self):
        '''
        Subclasses can override this to do something after values change.
        Default implementation calls 'child_value_changed' in the parent object if
        self._watched is True.
        '''
        if self._watched:
            self._mng.parent.child_value_changed(self)

    def set_default_value(self, value):
        self._default_value = value

    def revert_to_default(self):
        self.set(self._default_value)

    def set(self, value):
        if value == self._default_value:
            self._mng.del_value()
        else:
            self._mng.set_value(value)
        self.notify()

    def get(self):
        return self._mng.get_value(self._default_value)


class SessionValue(Value):
    '''
    This Value does not store itself in the value store.
    It will reset to its default value for each session.
    '''

    def set_default_value(self, value):
        super(SessionValue, self).set_default_value(value)
        self._value = value

    def set(self, value):
        self._value = value
        self._mng._on_value_changed()
        self.notify()

    def get(self):
        return self._value


class _TypedValue(Value):

    def validate(self, value):
        raise NotImplementedError()


class IntValue(_TypedValue):

    DEFAULT_EDITOR = 'int'

    def validate(self, value):
        if not value and value != 0:
            return self._default_value
        try:
            return int(value)
        except TypeError:
            raise TypeError(
                'Invalid value type %r for %s (must be a int)' % (
                    value, self.oid(),
                )
            )

    def set(self, value):
        value = self.validate(value)
        super(IntValue, self).set(value)

    def incr(self, by=1):
        self._mng.incr_value(by)
        self.notify()

    def decr(self, by=1):
        self._mng.decr_value(by)
        self.notify()


class BoolValue(_TypedValue):

    DEFAULT_EDITOR = 'bool'

    def validate(self, value):
        try:
            return bool(value)
        except TypeError:
            raise TypeError(
                'Invalid value type %r for %s (must be a bool)' % (
                    value, self.oid(),
                )
            )

    def set(self, value):
        value = self.validate(value)
        super(BoolValue, self).set(value)

    # deprecated, not needed (yet...)
    # def toggle(self):
    #     #FIXME: do this in the value store to be able to use redis pipeline !
    #     self.set(not self.get())


class FloatValue(_TypedValue):

    DEFAULT_EDITOR = 'float'

    def validate(self, value):
        if not value and value != 0:
            return self._default_value
        try:
            return float(value)
        except TypeError:
            raise TypeError(
                'Invalid value type %r for %s (must be a float)' % (
                    value, self.oid(),
                )
            )

    def set(self, value):
        super(FloatValue, self).set(self.validate(value))


class StringValue(_TypedValue):

    DEFAULT_EDITOR = 'string'

    def validate(self, value):
        try:
            return str(value)
        except TypeError:
            raise TypeError(
                'Invalid value type %r for %s (must be a str)' % (
                    value, self.oid(),
                )
            )

    def set(self, value):
        super(StringValue, self).set(self.validate(value))


class DictValue(_TypedValue):

    DEFAULT_EDITOR = 'mapping'

    def validate(self, value):
        try:
            return dict(value)
        except TypeError:
            raise TypeError(
                'Invalid value type %r for %s (must be a dict)' % (
                    value, self.oid(),
                )
            )

    def set(self, value):
        super(DictValue, self).set(self.validate(value))


class OrderedStringSetValue(Value):
    '''
    A list off string ordered by score.
    You cant set that value, you can only edit it.
    '''

    DEFAULT_EDITOR = 'set'

    def set_default_value(self, value):
        # We dont support default value but the Relation will give us one.
        # So the only acceptable value is None:
        if value is not None:
            raise Exception(
                'You cant configure the default of an OrderedStringSetValue')

    def revert_to_default(self):
        self._mng.del_value()

    def set(self, value):
        raise Exception('You cant set an OrderedStringSetValue')

    def get(self):
        return self._mng.oss_get()

    def get_range(self, first, last):
        return self._mng.oss_get_range(first, last)

    def has(self, member):
        has = self._mng.oss_has(member)
        self.notify()
        return has

    def add(self, member, score):
        self._mng.oss_add(member, score)
        self.notify()

    def remove(self, member):
        self._mng.oss_remove(member)
        self.notify()

    def len(self):
        return self._mng.oss_len()

    def get_score(self, member):
        return self._mng.oss_get_score(member)

    def set_score(self, member, score):
        self._mng.oss_set_score(member, score)
        self.notify()


class HashValue(_TypedValue):

    DEFAULT_EDITOR = 'mapping'

    def validate(self, value):
        if isinstance(value, dict):
            values = value.items()
        else:
            values = value
        try:
            return [(str(i), str(j)) for i, j in values]
        except:
            raise TypeError(
                'Invalid value type %r for %s (must be a mapping: dict or 2d list/tuple)' % (
                    value, self.oid(),
                )
            )

    def get_key(self, key):
        # hget
        return self._mng.hash_get_key(key)

    def has_key(self, key):
        # hexists
        return self._mng.hash_has_key(key)

    def del_key(self, key):
        # hdel
        self._mng.del_hash_key(key)
        self.notify()

    def as_dict(self):
        # hgetall
        return self._mng.get_hash_as_dict()

    def keys(self):
        # hkeys
        return self._mng.get_hash_keys()

    def len(self):
        # hlen
        return self._mng.get_hash_len()

    def update(self, **new_values):
        # hmset
        self._mng.update_hash(new_values)
        self.notify()

    def set_key(self, key, value):
        # hset
        self._mng.set_hash_key(key, value)
        self.notify()

    def set(self, value):
        value = self.validate(value)
        self._mng.set_hash(value)
        self.notify()

    def get(self):
        return self._mng.get_hash()


class ComputedValue(Value):

    def __init__(self, parent, name):
        super(ComputedValue, self).__init__(parent, name)
        self._store_value = False
        self._cached = False
        self._value = None
        self._dirty = True

    def set_cached(self, b):
        self._cached = bool(b)

    def set_store_value(self, b):
        self._store_value = bool(b)

    def touch(self):
        self._dirty = True
        super(ComputedValue, self).touch()

    def compute(self):
        '''
        Subclasses can override this to compute the value (and call set()).
        Default implementation calls 'compute_child_value' in the parent object.
        '''
        self._mng.parent.compute_child_value(self)

    def get(self):
        if self._dirty:
            self.compute()
        return self._value

    def set(self, value):
        if self._store_value:
            super(ComputedValue, self).set(value)
        self._value = value
        if self._cached:
            self._dirty = False
        self.notify()


class ChoiceValue(Value):
    '''
    This Value provides a list of potential/acceptable values.
    You can set the CHOICE class attribute to define this list.

    If STRICT_CHOICES is True, the value must exists in the CHOICES attribute
    or a ValueError will be raised by set()
    If STRICT_CHOICES is False, any value can do.

    '''
    DEFAULT_EDITOR = 'choice'

    STRICT_CHOICES = True
    CHOICES = []

    def choices(self):
        return self.__class__.CHOICES

    def set(self, value):
        if self.STRICT_CHOICES and value not in self.choices():
            # we still touch ourself, so we sure GUI refresh with the unchanged value:
            self.touch()
            raise ValueError('Invalid value %r. Should be one of %r' %
                             (value, self.choices()))
        super(ChoiceValue, self).set(value)


class MultiChoiceValue(ChoiceValue):
    '''
    A MultiChoiceValue is like a Choice but stores a list of those acceptable values.
    '''
    DEFAULT_EDITOR = 'multichoice'

    def add(self, extra_choice):
        self.set(self.get()+[extra_choice])

    def set(self, value):
        if self.STRICT_CHOICES:
            if not isinstance(value, list):
                value = [value]
            for v in value:
                if v not in self.choices():
                    raise ValueError(
                        'Invalid value %r. Should be one of %r' % (v, self.choices()))
        super(ChoiceValue, self).set(value)


class Ref(Value):
    '''
    A Ref stores a reference to an Object.
    The set() method accepts only Object of the type (or list of types) in SOURCE_TYPE class attribute.
    The get() method returns the Object.
    If you only need the Object's oid, use get_source_oid() as it does not require object lookup.
    '''

    ICON = 'ref'
    DEFAULT_EDITOR = 'ref'

    SOURCE_TYPE = None  # class_or_type_or_tuple

    @staticmethod
    def resolve_refs(object):
        max = 10
        nb = 0
        while isinstance(object, Ref):
            object = object.get()
            nb += 1
            if nb > max:
                raise RuntimeError('Maximum recursion exceeded')
        return object

    def __init__(self, parent, name):
        super(Ref, self).__init__(parent, name)
        self.source_object = None
        self._allow_cross_project = False
    
    def set_allow_cross_project(self, b):
        self._allow_cross_project = b

    def get(self):
        source_oid = self.get_source_oid()
        if self.source_object is None or self.source_object.oid() != source_oid:
            if source_oid is None:
                return None
            try:
                self.source_object = self._mng.get_object(source_oid)
            except (MissingChildError, MissingRelationError):
                raise RefSourceError(self.oid(), source_oid)

        return self.source_object

    def get_source_oid(self):
        return super(Ref, self).get()

    def can_set(self, source_object):
        try:
            self._assert_source_compatible(source_object)
        except ValueError:
            return False

        source_object = self.resolve_refs(source_object)

        try:
            self._validate_source_object(source_object)
        except RefSourceTypeError:
            return False
        else:
            return True

    def _assert_source_compatible(self, object):
        if not self._allow_cross_project and object.root() is not self.root():
            raise ValueError('Cannot connect to another flow!')

    def _validate_source_object(self, source_object):
        if source_object is not None:
            if self.SOURCE_TYPE is not None and not isinstance(source_object, self.SOURCE_TYPE):
                if 0:
                    # was too geeky :/
                    raise RefSourceTypeError(
                        'Ref %s cannot point to %r (should be a %r, but is a %s)' % (
                            self.oid(), source_object.oid(), self.SOURCE_TYPE, source_object.__class__
                        )
                    )
                else:
                    raise RefSourceTypeError(
                        'This is not a %s' % (self.SOURCE_TYPE.__name__,))
        return source_object

    def set(self, new_source_object):

        if new_source_object is not None:
            self._assert_source_compatible(new_source_object)
            new_source_object = self.resolve_refs(new_source_object)
            new_source_object = self._validate_source_object(new_source_object)
            new_oid = new_source_object.oid()
        else:
            new_oid = None

        old_source_oid = self.get_source_oid()
        if old_source_oid == new_oid:
            return

        super(Ref, self).set(new_oid)
        if old_source_oid is not None:
            try:
                old_source_object = self._mng.get_object(old_source_oid)
            except (MissingChildError, MissingRelationError):
                pass
            else:
                old_source_object._mng.remove_ref(self)

        self.source_object = new_source_object
        if self.source_object is not None:
            self.source_object._mng.add_ref(self)

#---OLD
    def source_touched(self):
        raise WIPException('I dont think this is still in use.')
        #print('++SOURCE TOUCHED', self.oid(), self.source_object.oid())
        self.touch()  # will touch my parent.

#---OLD
    def source_value_changed(self, old_value, new_value):
        raise WIPException('I think this is obsolete')
        #print('----->source_value_changed', self.oid(), old_value, new_value)
        self._mng._on_value_changed(old_value, new_value)



import six
from qtpy import QtWidgets, QtGui, QtCore
import qtpy
import functools

from kabaret.app import resources
from ..flow_layout import FlowLayout
from ..popup_menu import PopupMenu


class NavigationButton(QtWidgets.QToolButton):

    def __init__(self, name, oid, nav_widget):
        super(NavigationButton, self).__init__(nav_widget)
        self.name = name
        self.oid = oid
        self.nav_widget = nav_widget
        self._last_click_pos = QtCore.QPoint()

        self.setFont(self.nav_widget.bt_font)
        self.setProperty('tight_layout', True)
        self.setProperty('hide_arrow', True)
        self.setProperty('no_border', True)
        self.setProperty('square', True)
        self.setArrowType(QtCore.Qt.NoArrow)
        # self.setContextMenuPolicy(QtCore.Qt.ActionsContextMenu)

        self.setText('/%s' % (self.name,))

        self.clicked.connect(self._goto)

        self._menu = PopupMenu(self)

    def _goto_oid(self, oid):
        self.nav_widget._goto(oid)

    def _goto(self, b=None):
        self.nav_widget._goto(self.oid)

    def _show_navigables_menu(self, full_oid):
        self._menu.clear()

        m = self._menu

        m.addAction('Loading...')
        # self.setMenu(m)
        # self.showMenu()

        session = self.nav_widget._navigator.session
        try:
            navigatable_entries = session.cmds.Flow.get_navigable_oids(
                self.oid, full_oid
            )
        except Exception as err:
            m.clear()
            m.addAction('ERROR: ' + str(err))
            raise

        m.clear()
        root = m
        for i, entry in enumerate(navigatable_entries, 1):
            if entry is None:
                m.addMenu()
            else:
                label, oid = entry
                if oid == self.oid:
                    item = m.addItem(
                        '> %s <' % label
                    )
                    font = item.font()
                    font.setBold(True)
                    item.setFont(font)
                else:
                    m.addAction(label, callback=functools.partial(self._goto_oid, oid))

        root.addMenu()
        root.addAction("Copy", icon=resources.get_icon(('icons.gui', 'copy-document')),
                       callback=lambda: QtWidgets.QApplication.clipboard().setText(self.oid))
        root.addAction("Paste", icon=resources.get_icon(('icons.gui', 'paste-from-clipboard')),
                       callback=lambda: self._goto_oid(QtWidgets.QApplication.clipboard().text()))
        m.popup(QtGui.QCursor.pos())

    def mousePressEvent(self, e):
        if e.button() == QtCore.Qt.RightButton:
            self._show_navigables_menu(self.nav_widget.current_oid())
        else:
            self._last_click_pos = e.pos()
        super(NavigationButton, self).mousePressEvent(e)

    def mouseMoveEvent(self, e):
        super(NavigationButton, self).mouseMoveEvent(e)

        if self._last_click_pos.isNull():
            return
        drag_distance = (e.pos() - self._last_click_pos).manhattanLength()
        if drag_distance < QtWidgets.QApplication.startDragDistance():
            return

        oids = [self.oid]
        mime_data = QtCore.QMimeData()
        md = self.nav_widget._navigator.session.cmds.Flow.to_mime_data(oids)
        for data_type, data in six.iteritems(md):
            mime_data.setData(data_type, data)

        pixmap = QtGui.QPixmap(self.size())
        self.render(pixmap)

        # below makes the pixmap half transparent
        painter = QtGui.QPainter(pixmap)
        painter.setCompositionMode(painter.CompositionMode_DestinationIn)
        painter.fillRect(pixmap.rect(), QtGui.QColor(0, 0, 0, 127))
        painter.end()

        # make a QDrag
        drag = QtGui.QDrag(self)
        drag.setMimeData(mime_data)
        drag.setPixmap(pixmap)

        # shift the Pixmap so that it coincides with the cursor position
        drag.setHotSpot(self._last_click_pos)

        # start the drag operation
        # exec_ will return the accepted action from dropEvent
        drag_result = drag.exec_(QtCore.Qt.CopyAction)

    def mouseReleaseEvent(self, e):
        self._last_click_pos = QtCore.QPoint()
        super(NavigationButton, self).mouseReleaseEvent(e)


class NavigationOIDControls(QtWidgets.QWidget):

    def __init__(self, parent, navigator):
        super(NavigationOIDControls, self).__init__(parent)

        self._navigator = navigator

        self.bt_font = self.font()
        self.bt_font.setPointSize(self.bt_font.pointSize() * 1.3)

        self._flow_lo = FlowLayout()
        self.setLayout(self._flow_lo)
        self._flow_lo.setContentsMargins(2, 0, 2, 0)
        self._flow_lo.setSpacing(0)

    def update(self):
        self._flow_lo.clear()

        label_to_oid = self._navigator.split_current_oid()
        for label, goto_oid in label_to_oid:
            tb = NavigationButton(label, goto_oid, self)
            tb.adjustSize()
            self._flow_lo.addWidget(tb)

    def _goto_home(self):
        self._navigator.goto_root()

    def _goto(self, oid):
        self.setFocus()
        in_new_view = (
            QtWidgets.QApplication.keyboardModifiers() == QtCore.Qt.ControlModifier
        )
        self._navigator.goto(oid, in_new_view)

    def current_oid(self):
        return self._navigator.current_oid()


class NavigationHistoryControls(QtWidgets.QWidget):

    def __init__(self, parent, navigator):
        super(NavigationHistoryControls, self).__init__(parent)

        self._navigator = navigator

        self.prev_bt = QtWidgets.QToolButton(self)
        self.prev_bt.setProperty('no_border', True)
        self.prev_bt.setText('<')
        self.prev_bt.setIcon(resources.get_icon(
            ('icons.gui', 'chevron-sign-left'),
            disabled_ref=('icons.gui', 'chevron-sign-left-disabled')
        ))
        self.prev_bt.clicked.connect(self._on_prev_bt)

        self.up_bt = QtWidgets.QToolButton(self)
        self.up_bt.setProperty('no_border', True)
        self.up_bt.setText('/\\')
        self.up_bt.setIcon(resources.get_icon(
            ('icons.gui', 'chevron-sign-up'),
            disabled_ref=('icons.gui', 'chevron-sign-up-disabled')
        ))
        self.up_bt.clicked.connect(self._on_up_bt)

        self.next_bt = QtWidgets.QToolButton(self)
        self.next_bt.setProperty('no_border', True)
        self.next_bt.setText('>')
        self.next_bt.setIcon(resources.get_icon(
            ('icons.gui', 'chevron-sign-to-right'),
            disabled_ref=('icons.gui', 'chevron-sign-to-right-disabled')
        ))
        self.next_bt.clicked.connect(self._on_next_bt)

        self.home_bt = QtWidgets.QToolButton(self)
        self.home_bt.setProperty('no_border', True)
        self.home_bt.setText('/')
        self.home_bt.setIcon(resources.get_icon(
            ('icons.gui', 'home'),
            disabled_ref=('icons.gui', 'home-outline')
        ))
        self.home_bt.clicked.connect(self._goto_home)

        bt_lo = QtWidgets.QHBoxLayout()
        bt_lo.setContentsMargins(12, 0, 12, 0)
        bt_lo.setSpacing(0)
        bt_lo.addWidget(self.prev_bt)
        bt_lo.addWidget(self.up_bt)
        bt_lo.addWidget(self.next_bt)
        bt_lo.addWidget(self.home_bt)

        self.setLayout(bt_lo)

    def _goto_home(self):
        self._navigator.goto_root()

    def _on_prev_bt(self):
        # TODO: handle optional new view
        self._navigator.goto_prev()

    def _on_up_bt(self):
        # TODO: handle optional new view
        self._navigator.goto_parent()

    def _on_next_bt(self):
        # TODO: handle optional new view
        self._navigator.goto_next()

    def update(self):
        self.prev_bt.setEnabled(self._navigator.has_prev())
        self.up_bt.setEnabled(self._navigator.has_parent())
        self.next_bt.setEnabled(self._navigator.has_next())
        self.home_bt.setEnabled(self._navigator.current_oid is not None and self._navigator.current_oid() != '/Home')


class NavigationBar(QtWidgets.QWidget):

    def __init__(self, parent, navigator):
        super(NavigationBar, self).__init__(parent)
        layout = QtWidgets.QHBoxLayout(self)
        self._navigator = navigator
        self.nav_ctrl = NavigationHistoryControls(self, navigator)
        self.nav_oid = NavigationOIDControls(self, navigator)
        layout.addWidget(self.nav_ctrl, 10, alignment=QtCore.Qt.AlignVCenter)
        layout.addWidget(self.nav_oid, 90, alignment=QtCore.Qt.AlignVCenter)
        self.setLayout(layout)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        source = event.source()
        if source is not None:
            children = self.findChildren(type(source), source.objectName())
            if children and source in children:  # the drop come from one of its children
                return

        if event.mimeData().hasFormat("text/plain"):
            oid = event.mimeData().text()
            try:
                self._navigator.session.cmds.Flow.resolve_path(oid)
            except:
                pass
            else:
                event.acceptProposedAction()

    def dropEvent(self, event):
        oid = event.mimeData().text()
        oid = self._navigator.session.cmds.Flow.resolve_path(oid)
        self._navigator.goto(oid)
        event.acceptProposedAction()





from kabaret.flow import (
    values,
    Object, Map, Action, ConnectAction,
    ChoiceValueSetAction, ChoiceValueSelectAction,
    Child, Parent, Computed, Connection,
    Param, IntParam, BoolParam, HashParam
)

from .icon_browser import IconBrowserGroup
from .showcase import (
    ValuesGroup, 
    MapsGroup, 
    EditorsGroup, 
    RelationsGroup, 
    ActionsGroup, 
    UIConfigGroup,
    InjectionGroup,
    OtherFeaturesGroup,
)

from .showcase.custom_page import CustomPageGroup


class UnittestProject(Object):

    values = Child(ValuesGroup)
    maps = Child(MapsGroup)
    relations = Child(RelationsGroup)
    editors = Child(EditorsGroup)
    ui_config = Child(UIConfigGroup)
    actions = Child(ActionsGroup)
    custom_page = Child(CustomPageGroup)
    injection = Child(InjectionGroup)
    other = Child(OtherFeaturesGroup)
    icons = Child(IconBrowserGroup)



from kabaret import flow

class CheckSizeAction(flow.Action):

    _parent = flow.Parent()

    def allow_context(self, context):
        return context.endswith('.details')

    def needs_dialog(self):
        return True

    def get_buttons(self):
        size = self._parent.default_size.get()
        min_size = self._parent.minimum_size.get()
        max_size = self._parent.maximum_size.get()

        if size == min_size == max_size == 'disabled':
            msg = (
                "<p>This Action dialog has been resized according to its content "
                "since no option was provided.</p>"
            )
        else:
            msg = ""

            if size is not None and size != (-1, -1):
                if (min_size is not None and size < min_size) or (max_size is not None and size > max_size):
                    msg += (
                        "<p>The size of this Action dialog has been computed to "
                        "fit its content (out of provided range)."
                    )
                else:
                    msg += f"<p>The default size of this Action dialog is <b>{str(size)}</b>.</p>"
            else:
                msg += (
                    f"<p>The size of this Action dialog has been computed to "
                    "fit its content (default size).</p>"
                )
            if min_size is not None and min_size != (-1, -1):
                msg += f"<p>You cannot resize it below <b>{str(min_size)}</b>.</p>"
            if max_size is not None and max_size != (-1, -1):
                msg += f"<p>You cannot resize it beyond <b>{str(max_size)}</b>.</p>"
        
        self.message.set(f"<html>{msg}</html>")
        
        return ['Nice']

    def run(self, button):
        return None

    def _fill_ui(self, ui):
        size = self._parent.default_size.get()

        if size is not None:
            ui['dialog_size'] = size

        min_size = self._parent.minimum_size.get()

        if min_size is not None:
            ui['dialog_min_size'] = min_size

        max_size = self._parent.maximum_size.get()

        if max_size is not None:
            ui['dialog_max_size'] = max_size

class EditSizeAction(flow.Action):

    width = flow.SessionParam(0).ui(editor='int')
    height = flow.SessionParam(0).ui(editor='int')

    _size = flow.Parent()
    _options = flow.Parent(2)

    def allow_context(self, context):
        return context and context.endswith('.inline')
    
    def needs_dialog(self):
        size = self._size.get() or self._size._default_value
        self.width.set(size[0])
        self.height.set(size[1])
        
        return True

    def get_buttons(self):
        return ['Save', 'Reset', 'Cancel']
    
    def run(self, button):
        if button == 'Cancel':
            return
        
        if button == 'Reset':
            self._size.revert_to_default()
        else:
            self._size.set((self.width.get(), self.height.get()))

class DisableSizeAction(flow.Action):

    _size = flow.Parent()

    def allow_context(self, context):
        return context and context.endswith('.inline')
    
    def needs_dialog(self):
        return False
    
    def run(self, button):
        self._size.set(None)

class EditableSizeValue(flow.values.SessionValue):

    edit = flow.Child(EditSizeAction)
    disable = flow.Child(DisableSizeAction)

class DialogSizeOptions(flow.Object):

    doc = flow.Label(
        "You can tweak the different size options below and "
        "see their effects on the behaviour of the Action dialog."
    )
    
    default_size = flow.SessionParam((256, 256), EditableSizeValue).watched().ui(editable=False)
    minimum_size = flow.SessionParam((128, 128), EditableSizeValue).watched().ui(editable=False)
    maximum_size = flow.SessionParam((512, 512), EditableSizeValue).watched().ui(editable=False)

    code = flow.Computed().ui(editor='textarea', html=True, editable=False)

    check_size_result = flow.Child(CheckSizeAction)

    def child_value_changed(self, child_value):
        if child_value in (self.default_size, self.minimum_size, self.maximum_size):
            self.code.touch()

    def compute_child_value(self, child_value):
        if child_value is self.code:
            ui = {}
            if self.default_size.get() is not None:
                ui['dialog_size'] = self.default_size.get()
            if self.minimum_size.get() is not None:
                ui['dialog_min_size'] = self.minimum_size.get()
            if self.maximum_size.get() is not None:
                ui['dialog_max_size'] = self.maximum_size.get()
            
            code = "<pre>action = Child(Action)"
            if ui:
                code += ".ui(<br>{}<br>)".format(',<br>'.join([f'    {k}={v}' for k, v in ui.items()]))
            
            self.code.set(code + "</pre>")


from qtpy import QtWidgets


class AutoHeightTreeWidget(QtWidgets.QTreeWidget):

    def __init__(self, *args, **kwargs):
        self._min_h = 100
        super(AutoHeightTreeWidget, self).__init__(*args, **kwargs)

        # self.setSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Maximum )

    def updateAutoHeight(self):
        height = self.header().height() + 2

        for it in QtWidgets.QTreeWidgetItemIterator(self, QtWidgets.QTreeWidgetItemIterator.All):
            height += self.visualItemRect(it.value()).height()

        height = max(self._min_h, height)
        self.setFixedHeight(height)


class RefByWidget(QtWidgets.QWidget):

    def __init__(self, parent, oids, leaf_callback, group_callback, hide_callback=None):
        super(RefByWidget, self).__init__(parent)

        self.oids = []

        self._is_open = False

        self._open_bt = QtWidgets.QToolButton(self)
        self._open_bt.setText('Show References')
        self._open_bt.clicked.connect(self.toggle_open)

        self._is_compact = False

        self._compact_cb = QtWidgets.QCheckBox(self)
        self._compact_cb.setText('Compact')
        self._compact_cb.setChecked(self._is_compact)
        self._compact_cb.toggled.connect(self._on_compact_cb)

        self._tree = AutoHeightTreeWidget(self)
        self._tree.setColumnCount(1)
        self._tree.setHeaderLabels(['Connected to'])
        self._tree.itemDoubleClicked.connect(self._on_tree_dble_click)
        self._tree.hide()

        hlo = QtWidgets.QHBoxLayout()
        hlo.addWidget(self._open_bt)
        hlo.addWidget(self._compact_cb)
        hlo.addStretch(100)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addLayout(hlo)
        self.layout().addWidget(self._tree)

        self._leaf_callback = leaf_callback
        self._group_callback = group_callback
        self._hide_callback = hide_callback

        self.set(oids)

        self.open()

    def set_togglable(self, b):
        self._open_bt.setVisible(b)

    def is_open(self):
        return self._is_open

    def is_compact(self):
        return self._is_compact

    def set(self, oids):
        self.oids = oids
        if oids:
            self._open_bt.setText('Show %d References' % (len(self.oids),))
        else:
            self._open_bt.setText('No Reference.')
        self.refresh()

    def _load_tree(self):

        tree = dict(oid='', name='', children=[])
        oid_to_item = {'': tree}

        def add_item(oid):
            try:
                return oid_to_item[oid]
            except KeyError:
                pass

            me = dict(oid=oid, name=oid.rsplit('/', 1)[-1], children=[])
            oid_to_item[oid] = me
            parent_oid, name = oid.rsplit('/', 1)
            parent = add_item(parent_oid)
            parent['children'].append(me)
            return me

        for oid in self.oids:
            add_item(oid)

        def compact_leaves(item):
            children = item['children']
            if not children:
                return
            if len(children) == 1:
                child = children[0]
                if not child['children']:
                    item['oid'] = child['oid']
                    item['name'] += '.' + child['name']
                    item['children'] = []
            for child in item['children']:
                compact_leaves(child)

        compact_leaves(tree)

        if self._is_compact:
            def compact_items(item):
                children = item['children']
                if not children:
                    return

                if len(children) == 1:
                    child = children[0]
                    if child['children']:
                        item['oid'] = child['oid']
                        item['name'] += '/' + child['name']
                        item['children'] = child['children']
                        compact_items(item)
                else:
                    for child in item['children']:
                        compact_items(child)

            compact_items(tree)

        def create_tree_item(parent, item):
            oid = item['oid']
            if oid:
                name = item['name']
                tree_item = QtWidgets.QTreeWidgetItem(parent, [name])
                tree_item.flow_id = oid
                if item['children']:
                    tree_item.action = self._group_callback
                else:
                    tree_item.action = self._leaf_callback
            else:
                tree_item = parent
            for child in item['children']:
                create_tree_item(tree_item, child)

        create_tree_item(self._tree, tree)

        self._tree.expandAll()
        self._tree.updateAutoHeight()

    def clear(self):
        self._tree.clear()

    def refresh(self):
        self.clear()
        self._load_tree()

    def open(self):
        self._is_open = True
        self._tree.show()
        self._open_bt.setText('Hide References')
        self._compact_cb.show()
        self.refresh()
        if self._hide_callback:
            self._hide_callback(False)

    def close(self):
        self._is_open = False
        self._tree.hide()
        self._open_bt.setText('Show %s References' % (len(self.oids, )))
        self._compact_cb.hide()
        self.refresh()
        if self._hide_callback:
            self._hide_callback(True)

    def toggle_open(self):
        if self._is_open:
            self.close()
        else:
            self.open()

    def _on_compact_cb(self, b):
        self._is_compact = b and True or False
        self._tree.clear()
        self._load_tree()

    def set_compact(self, b):
        self._is_compact = b and True or False
        self._compact_cb.setChecked(b)
        self._tree.clear()
        self._load_tree()

    def _on_tree_dble_click(self, item, column):
        item.action(item.flow_id)



"""
    kabaret.flow.exceptions:
    all the exception raised by flow objects.
"""


class WIPException(Exception):
    '''
    Use while code update is in progress
    '''


class MissingRelationError(KeyError):

    def __init__(self, oid, relation_name):
        super(MissingRelationError, self).__init__("Could not find related object %r under %r" % (relation_name, oid))


class MissingChildError(KeyError):

    def __init__(self, oid, child_name):
        super(MissingChildError, self).__init__("Could not find child %r under %r" % (child_name, oid))
        self.oid = oid
        self.child_name = child_name


class MappedNameError(MissingChildError):

    def __init__(self, oid, mapped_name):
        super(MappedNameError, self).__init__(oid, mapped_name)
        self.message = "No mapped name %r in %r" % (mapped_name, oid)


class RefSourceTypeError(ValueError):
    pass


class RefSourceError(ValueError):

    def __init__(self, ref_oid, source_oid):
        super(RefSourceError, self).__init__("Could not find source %r for ref %r" % (source_oid, ref_oid))
        self.ref_oid = ref_oid
        self.source_oid = source_oid
